/*
 * vzct_conf.c: core driver functions for managing
 * Virtuozzo Containers
 *
 * Copyright (C) 2017 Virtuozzo, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>

#include "capabilities.h"
#include "viralloc.h"
#include "domain_conf.h"
#include "virlog.h"
#include "virclosecallbacks.h"
#include "virstring.h"
#include "configmake.h"

#include "vzct_utils.h"
#include "vzct_conf.h"
#include "vzct_vzctl.h"

#define VIR_FROM_THIS VIR_FROM_VZCT

#define VZCT_CONFIG_DIR SYSCONFDIR "/libvirt/vzct"
#define VZCT_AUTOSTART_DIR SYSCONFDIR "/libvirt/vzct-autostart"
#define VZCT_STATE_DIR "/vz/vzct"
#define VZCT_SAVE_DIR "/vz/vzsaves"
#define VZCT_SNAPSHOT_DIR "/vz/vzsnapshots"


VIR_LOG_INIT("vzct.vzct_conf");

static virClassPtr vzctDriverConfigClass;
static void vzctDriverConfigDispose(void *obj);

static int
vzctConfigOnceInit(void)
{
    if (!VIR_CLASS_NEW(vzctDriverConfig, virClassForObject()))
        return -1;

    return 0;
}

VIR_ONCE_GLOBAL_INIT(vzctConfig)

virCapsPtr
vzctDriverCapsInit(void)
{
    virCapsPtr caps;
    virCapsGuestPtr guest;

    if ((caps = virCapabilitiesNew(virArchFromHost(),
                                   false, false)) == NULL)
        goto error;

    if (virCapabilitiesInitNUMA(caps) < 0)
        goto error;

    if (virGetHostUUID(caps->host.host_uuid)) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       "%s", _("cannot get the host uuid"));
        goto error;
    }

    if ((guest = virCapabilitiesAddGuest(caps,
                                         VIR_DOMAIN_OSTYPE_EXE,
                                         caps->host.arch,
                                         NULL,
                                         NULL,
                                         0,
                                         NULL)) == NULL)
        goto error;

    if (virCapabilitiesAddGuestDomain(guest,
                                      VIR_DOMAIN_VIRT_VZCT,
                                      NULL,
                                      NULL,
                                      0,
                                      NULL) == NULL)
        goto error;

    return caps;

 error:
    virObjectUnref(caps);
    return NULL;
}


virCapsPtr
vzctDriverGetCapabilities(vzctDriverPtr driver)
{
    return virObjectRef(driver->caps);
}

static void *
vzctDomainObjPrivateAlloc(void* opaque ATTRIBUTE_UNUSED)
{
    vzctDomainObjPrivatePtr priv;

    VIR_DEBUG("Private alloc");

    if (VIR_ALLOC(priv) < 0)
        return NULL;

    if (virCondInit(&priv->jobCond) < 0) {
        VIR_FREE(priv);
        return NULL;
    }

    priv->vncPid = -1;
    priv->migrationPid = -1;

    return priv;
}


static void
vzctDomainObjPrivateFree(void *data)
{
    vzctDomainObjPrivatePtr priv = data;

    VIR_DEBUG("Private free");

    virCondDestroy(&priv->jobCond);
    VIR_FREE(priv);
}

virDomainXMLPrivateDataCallbacks vzctDriverPrivateDataCallbacks = {
    .alloc = vzctDomainObjPrivateAlloc,
    .free = vzctDomainObjPrivateFree,
};

static int vzctGetMaxNetworkIdx(const virDomainDef *def)
{
    int max_idx = -1, idx;
    size_t i;

    for (i = 0; i<def->nnets; i++) {
        const char *c;
        virDomainNetDefPtr net = def->nets[i];
        if (!net->ifname_guest)
            continue;
        if (STRNEQLEN("eth", net->ifname_guest, 3))
            continue;
        c = net->ifname_guest + 3;
        idx = virParseNumber(&c);
        if (max_idx < idx)
            max_idx = idx;
    }
    return max_idx;
}

static int
vzctDomainDeviceDefPostParse(virDomainDeviceDefPtr dev,
                             const virDomainDef *def,
                             virCapsPtr caps ATTRIBUTE_UNUSED,
                             unsigned int parseFlags ATTRIBUTE_UNUSED,
                             void *opaque ATTRIBUTE_UNUSED,
                             void *parseOpaque ATTRIBUTE_UNUSED)
{
    VIR_DEBUG("post-parse-device %s", def->name);
    if (dev->type == VIR_DOMAIN_DEVICE_NET) {
        /* device name in guest is required */
        virDomainNetDefPtr net = dev->data.net;
        if (!net->ifname_guest) {
            int idx = vzctGetMaxNetworkIdx(def);
            if (!virAsprintf(&net->ifname_guest, "eth%d", idx + 1))
                return -1;
        }

        if (!net->ifname_guest_actual &&
            VIR_STRDUP(net->ifname_guest_actual, net->ifname_guest) < 0)
            return -1;

    } else if (dev->type == VIR_DOMAIN_DEVICE_DISK) {
        virDomainDiskDefPtr disk = dev->data.disk;

        if (!virDomainDiskGetDriver(disk) &&
            virDomainDiskSetDriver(disk, "ploop") < 0)
            return -1;

        if (virDomainDiskGetFormat(disk) == VIR_STORAGE_FILE_NONE) {
            if (virDomainDiskGetType(disk) == VIR_STORAGE_TYPE_FILE)
                virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_PLOOP);
            else if (virDomainDiskGetFormat(disk) == VIR_STORAGE_FILE_NONE)
                virDomainDiskSetFormat(disk, VIR_STORAGE_FILE_RAW);
        }
    } else if (dev->type == VIR_DOMAIN_DEVICE_FS) {
        virDomainFSDefPtr fs = dev->data.fs;

        if (fs->fsdriver == VIR_DOMAIN_FS_DRIVER_TYPE_DEFAULT)
            fs->fsdriver = VIR_DOMAIN_FS_DRIVER_TYPE_PLOOP;

        if (fs->format == VIR_STORAGE_FILE_NONE) {
            if (fs->type == VIR_DOMAIN_FS_TYPE_FILE)
                fs->format = VIR_STORAGE_FILE_PLOOP;
            else if (fs->type == VIR_DOMAIN_FS_TYPE_BLOCK)
                fs->format = VIR_STORAGE_FILE_RAW;
        }
    }

    return 0;
}

static int
vzctValidateGraphics(const virDomainGraphicsDef *gr)
{
    if (gr->type != VIR_DOMAIN_GRAPHICS_TYPE_VNC) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only VNC graphics is supported."));
        return -1;
    }

    if (gr->data.vnc.websocket != 0) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Websocket VNC ports are not supported."));
        return -1;
    }

    if (gr->data.vnc.keymap != 0 &&
        STRNEQ(gr->data.vnc.keymap, "en-us")) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only 'en-us' VNC keymap is supported."));
        return -1;
    }

    if (gr->data.vnc.sharePolicy == VIR_DOMAIN_GRAPHICS_VNC_SHARE_ALLOW_EXCLUSIVE) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("VNC exclusive share policy is not supported."));
        return -1;
    }

    if (gr->data.vnc.auth.connected == VIR_DOMAIN_GRAPHICS_AUTH_CONNECTED_FAIL ||
        gr->data.vnc.auth.connected == VIR_DOMAIN_GRAPHICS_AUTH_CONNECTED_KEEP) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("VNC supports only disconnecton on passwd change."));
        return -1;
    }

    if (gr->data.vnc.auth.expires) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("VNC expiring passwords is not supported."));
        return -1;
    }

    if (gr->nListens > 1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only one graphics listen address is supported."));
        return -1;
    }

    if (gr->listens[0].type != VIR_DOMAIN_GRAPHICS_LISTEN_TYPE_ADDRESS) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only listening on inet address is supported for graphics"));
        return -1;
    }

    return 0;
}

static int
vzctDomainDefValidateCallback(const virDomainDef *def,
                              virCapsPtr caps ATTRIBUTE_UNUSED,
                              void *opaque ATTRIBUTE_UNUSED)
{
    size_t i;
    int diskIndex = 0;
    int netIndex = 0;

    for (i = 0; i < def->ndisks; i++) {
        int idx = def->disks[i]->info.bootIndex;

        if (idx && (!diskIndex || idx < diskIndex))
            diskIndex = idx;
    }

    for (i = 0; i < def->nnets; i++) {
        int idx = def->nets[i]->info.bootIndex;

        if (idx && (!netIndex || idx < netIndex))
            netIndex = idx;
    }

    if (netIndex && (!diskIndex || netIndex < diskIndex)) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Booting from network interface is unsupported."));
        return -1;
    }

    if (def->ngraphics > 1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only one graphics per domain is supported."));
        return -1;
    }

    if (def->placement_mode != VIR_DOMAIN_CPU_PLACEMENT_MODE_STATIC) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Only one static vcpu pinning is supported."));
        return -1;
    }

    return 0;
}

static int
vzctValidateNet(const virDomainNetDef *net)
{
    switch (net->type) {
    case VIR_DOMAIN_NET_TYPE_BRIDGE:
    case VIR_DOMAIN_NET_TYPE_NETWORK:
    case VIR_DOMAIN_NET_TYPE_ETHERNET:
    case VIR_DOMAIN_NET_TYPE_DIRECT:
        return 0;

    case VIR_DOMAIN_NET_TYPE_USER:
    case VIR_DOMAIN_NET_TYPE_VHOSTUSER:
    case VIR_DOMAIN_NET_TYPE_SERVER:
    case VIR_DOMAIN_NET_TYPE_CLIENT:
    case VIR_DOMAIN_NET_TYPE_MCAST:
    case VIR_DOMAIN_NET_TYPE_UDP:
    case VIR_DOMAIN_NET_TYPE_INTERNAL:
    case VIR_DOMAIN_NET_TYPE_LAST:
    case VIR_DOMAIN_NET_TYPE_HOSTDEV:
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Network device type is not supported"));
        break;
    }
    return -1;
}

static int
vzctDeviceDefValidateCallback(const virDomainDeviceDef *dev,
                              const virDomainDef *def,
                              void *opaque ATTRIBUTE_UNUSED)
{
    virDomainFSDefPtr fs;
    virDomainDiskDefPtr disk;
    VIR_DEBUG("validate-device %s", def->name);

    switch (dev->type) {
    case VIR_DOMAIN_DEVICE_FS:
        fs = dev->data.fs;
        if (fs->type != VIR_DOMAIN_FS_TYPE_FILE &&
            fs->type != VIR_DOMAIN_FS_TYPE_BLOCK) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only files and blockdevs are supported "
                             "for filesystems"));
            return -1;
        }

        if (fs->type == VIR_DOMAIN_FS_TYPE_FILE &&
            fs->format != VIR_STORAGE_FILE_PLOOP) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only ploop files are supported for filesystems"));
            return -1;
        }

        if (fs->type == VIR_DOMAIN_FS_TYPE_BLOCK &&
            fs->format != VIR_STORAGE_FILE_RAW) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only raw blockdevs are supported for filesystems"));
            return -1;
        }

        if (fs->fsdriver != VIR_DOMAIN_FS_DRIVER_TYPE_PLOOP) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only ploop driver is supported for filesystems"));
            return -1;
        }

        return 0;

    case VIR_DOMAIN_DEVICE_DISK:
        disk = dev->data.disk;
        if (disk->src->type != VIR_STORAGE_TYPE_FILE &&
            disk->src->type != VIR_STORAGE_TYPE_BLOCK) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only files and blockdevs are supported "
                             "for disks"));
            return -1;
        }

        if (disk->src->type == VIR_STORAGE_TYPE_FILE &&
            disk->src->format != VIR_STORAGE_FILE_PLOOP) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only ploop files are supported for disks"));
            return -1;
        }

        if (disk->src->type == VIR_STORAGE_TYPE_BLOCK &&
            disk->src->format != VIR_STORAGE_FILE_RAW) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only raw blockdevs are supported for disk"));
            return -1;
        }

        if (STRNEQ(virDomainDiskGetDriver(disk), "ploop")) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only ploop driver is supported for disks"));
            return -1;
        }

        return 0;

    case VIR_DOMAIN_DEVICE_NET:
        return vzctValidateNet(dev->data.net);

    case VIR_DOMAIN_DEVICE_GRAPHICS:
        return vzctValidateGraphics(dev->data.graphics);

    case VIR_DOMAIN_DEVICE_CONTROLLER:
    case VIR_DOMAIN_DEVICE_VIDEO:
    case VIR_DOMAIN_DEVICE_CHR:
        return 0;
    default:
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED,
                       _("Virtuozzo containers don't support devices of type '%s'."),
                       virDomainDeviceTypeToString(dev->type));
        return -1;
    }
    return 0;
}

virDomainDefParserConfig vzctDriverDomainDefParserConfig = {
    .devicesPostParseCallback = vzctDomainDeviceDefPostParse,
    .domainValidateCallback = vzctDomainDefValidateCallback,
    .deviceValidateCallback = vzctDeviceDefValidateCallback,
};


virDomainXMLOptionPtr
vzctDomainXMLConfInit(void)
{
    return virDomainXMLOptionNew(&vzctDriverDomainDefParserConfig,
                                 &vzctDriverPrivateDataCallbacks,
                                 NULL, NULL, NULL);
}

vzctDriverConfigPtr
vzctDriverConfigNew(void)
{
    vzctDriverConfigPtr cfg;

    if (vzctConfigInitialize() < 0)
        return NULL;

    if (!(cfg = virObjectNew(vzctDriverConfigClass)))
        return NULL;

    if (VIR_STRDUP(cfg->configDir, VZCT_CONFIG_DIR) < 0)
        goto error;
    if (VIR_STRDUP(cfg->stateDir, VZCT_STATE_DIR) < 0)
        goto error;
    if (VIR_STRDUP(cfg->saveDir, VZCT_SAVE_DIR) < 0)
        goto error;
    if (VIR_STRDUP(cfg->autostartDir, VZCT_AUTOSTART_DIR) < 0)
        goto error;
    if (VIR_STRDUP(cfg->snapshotDir, VZCT_SNAPSHOT_DIR) < 0)
        goto error;


    return cfg;
 error:
    virObjectUnref(cfg);
    return NULL;
}

int
vzctLoadDriverConfig(vzctDriverConfigPtr cfg ATTRIBUTE_UNUSED,
                     const char *filename)
{
    virConfPtr conf;
    int ret = -1;

    /* Avoid error from non-existent or unreadable file. */
    if (access(filename, R_OK) == -1)
        return 0;

    conf = virConfReadFile(filename, 0);
    if (!conf)
        return -1;

    ret = 0;

    virConfFree(conf);
    return ret;
}

vzctDriverConfigPtr
vzctDriverGetConfig(vzctDriverPtr driver)
{
    vzctDriverConfigPtr cfg;
    vzctDriverLock(driver);
    cfg = virObjectRef(driver->config);
    vzctDriverUnlock(driver);
    return cfg;
}

static void
vzctDriverConfigDispose(void *obj)
{
    vzctDriverConfigPtr cfg = obj;

    VIR_FREE(cfg->configDir);
    VIR_FREE(cfg->autostartDir);
    VIR_FREE(cfg->stateDir);
    VIR_FREE(cfg->saveDir);
    VIR_FREE(cfg->snapshotDir);
}
