/*
 * vzct_driver.c: core driver functions for managing
 * Virtuozzo Containers
 *
 * Copyright (C) 2017 Virtuozzo, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>
#include <signal.h>
#include <linux/netlink.h>

#include <vzctl/libvzctl.h>

#include "virlog.h"
#include "viraccessapicheck.h"
#include "viralloc.h"
#include "virhostcpu.h"
#include "virhostmem.h"
#include "configmake.h"
#include "virfile.h"
#include "vircommand.h"
#include "virstring.h"
#include "virpidfile.h"
#include "virprocess.h"
#include "domain_audit.h"
#include "virdomainobjlist.h"
#include "virnetdevtap.h"
#include "virtime.h"
#include "viratomic.h"
#include "util/virnetdevexe.h"
#include "cpu/cpu.h"

#include "vzct_driver.h"
#include "vzct_utils.h"
#include "vzct_conf.h"
#include "vzct_vzctl.h"

static vzctDriverPtr vzct_driver;

#define VIR_FROM_THIS VIR_FROM_VZCT

VIR_LOG_INIT("vzct.vzct_driver");

typedef enum {
    VZCT_DOMAIN_JOB_SHUTDOWN,
    VZCT_DOMAIN_JOB_LAST
} vzctDomainJobType;

typedef struct _vzctDomainJob vzctDomainJob;
typedef vzctDomainJob *vzctDomainJobPtr;

struct _vzctDomainJob {
    vzctDomainJobType type;
    virDomainObjPtr vm;
};

static void
vzctStopMonitor(vzctDriverPtr driver);
static void
vzctDomainStop(vzctDriverPtr driver,
               virDomainObjPtr vm,
               virDomainShutoffReason reason);

static int
vzctDomainObjBeginJob(virDomainObjPtr vm, bool *jobStarted)
{
    vzctDomainObjPrivatePtr priv = vm->privateData;
    unsigned long long now;
    unsigned long long then;

    if (virTimeMillisNow(&now) < 0)
        return -1;

    then = now + 30 * 1000;

    while (priv->job) {
        if (virObjectWaitUntil(vm, &priv->jobCond, then) < 0) {
            if (errno == ETIMEDOUT)
                virReportError(VIR_ERR_OPERATION_TIMEOUT, "%s",
                               _("cannot acquire job"));
            else
                virReportSystemError(errno, "%s",
                                     _("cannot acquire job mutex"));

            return -1;
        }
    }

    if (virTimeMillisNow(&now) < 0)
        return -1;

    priv->job = true;
    priv->jobStarted = now;
    priv->migration = false;

    if (jobStarted)
        *jobStarted = true;

    return 0;
}

static void
vzctDomainObjEndJob(virDomainObjPtr vm, bool *jobStarted)
{
    vzctDomainObjPrivatePtr priv = vm->privateData;

    if (jobStarted && !*jobStarted)
        return;

    priv->job = false;
    virCondSignal(&priv->jobCond);
}

static virDomainObjPtr
vzctDomObjFromDomain(virDomainPtr domain)
{
    virDomainObjPtr vm;
    vzctDriverPtr driver = domain->conn->privateData;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    vm = virDomainObjListFindByUUIDRef(driver->domains, domain->uuid);
    if (!vm) {
        virUUIDFormat(domain->uuid, uuidstr);
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("no domain with matching uuid '%s' (%s)"),
                       uuidstr, domain->name);
        return NULL;
    }

    return vm;
}

static virDrvOpenStatus
vzctConnectOpen(virConnectPtr conn,
                virConnectAuthPtr auth ATTRIBUTE_UNUSED,
                virConfPtr conf ATTRIBUTE_UNUSED,
                unsigned int flags)
{
    virCheckFlags(VIR_CONNECT_RO, VIR_DRV_OPEN_ERROR);

    if (!conn->uri)
        return VIR_DRV_OPEN_DECLINED;

    if (!conn->uri->scheme)
        return VIR_DRV_OPEN_DECLINED;

    if (STRNEQ(conn->uri->scheme, "vzct"))
        return VIR_DRV_OPEN_DECLINED;

    /* Remote driver should handle these. */
    if (conn->uri->server)
        return VIR_DRV_OPEN_DECLINED;

    /* From this point on, the connection is for us. */
    if (STRNEQ_NULLABLE(conn->uri->path, "/system")) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Unexpected Virtuozzo URI path '%s', try vzct:///system"),
                       conn->uri->path);
        return VIR_DRV_OPEN_ERROR;
    }

    if (virConnectOpenEnsureACL(conn) < 0)
        return VIR_DRV_OPEN_ERROR;

    if (!vzct_driver) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       "%s", _("vzct state driver is not active"));
        return VIR_DRV_OPEN_ERROR;
    }

    conn->privateData = vzct_driver;

    return VIR_DRV_OPEN_SUCCESS;
}

static int
vzctConnectClose(virConnectPtr conn)
{
    virCloseCallbacksRun(vzct_driver->closeCallbacks, conn, vzct_driver->domains, vzct_driver);
    conn->privateData = NULL;

    return 0;
}

static int
vzctStateCleanup(void)
{
    if (!vzct_driver)
        return 0;

    virThreadPoolFree(vzct_driver->workerPool);
    vzctStopMonitor(vzct_driver);
    virObjectUnref(vzct_driver->domains);
    virObjectUnref(vzct_driver->domainEventState);
    virPortAllocatorRangeFree(vzct_driver->vncPorts);
    virPortAllocatorRangeFree(vzct_driver->migrationPorts);
    virObjectUnref(vzct_driver->closeCallbacks);
    virSysinfoDefFree(vzct_driver->hostsysinfo);
    virObjectUnref(vzct_driver->caps);
    virObjectUnref(vzct_driver->xmlopt);
    virObjectUnref(vzct_driver->config);
    virMutexDestroy(&vzct_driver->lock);
    VIR_FREE(vzct_driver);
    vzct_driver = NULL;
    return 0;
}


static int
vzctCheckState(virDomainObjPtr dom, void *opaque)
{
    vzctDriverPtr driver = opaque;
    vzctDomainObjPrivatePtr priv = dom->privateData;
    virDomainDefPtr def = dom->def;
    char *pidfile = NULL;
    int ret = -1;

    virObjectRef(dom);
    virObjectLock(dom);

    if (!virDomainObjIsActive(dom)) {
        ret = 0;
        goto cleanup;
    }

    if (def->ngraphics > 0) {
        if (virAsprintf(&pidfile, "/vz/vzct/%s.vnc.pid", def->name) < 0)
            goto cleanup;

        if (virPidFileReadPath(pidfile, &priv->vncPid) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Failed to read vnc server pid for %s"), def->name);
            goto cleanup;
        }
    }

    if (!vzctlDomainActive(dom))
        goto cleanup;

    ret = 0;

 cleanup:
    if (ret < 0)
        vzctDomainStop(driver, dom, VIR_DOMAIN_SHUTOFF_UNKNOWN);
    VIR_FREE(pidfile);
    virDomainObjEndAPI(&dom);

    /* return success anyway, so we don't fail daemon start on
     * single domain failure */
    return 0;
}

static int
vzctDomainSnapshotLoad(virDomainObjPtr vm,
                       void *data)
{
    char *baseDir = (char *)data;
    char *snapDir = NULL;
    DIR *dir = NULL;
    struct dirent *entry;
    char *xmlStr;
    char *fullpath;
    virDomainSnapshotDefPtr def = NULL;
    virDomainSnapshotObjPtr snap = NULL;
    virDomainSnapshotObjPtr current = NULL;
    unsigned int flags = (VIR_DOMAIN_SNAPSHOT_PARSE_REDEFINE |
                          VIR_DOMAIN_SNAPSHOT_PARSE_DISKS |
                          VIR_DOMAIN_SNAPSHOT_PARSE_INTERNAL);
    int ret = -1;
    virCapsPtr caps = NULL;
    int direrr;

    virObjectLock(vm);
    if (virAsprintf(&snapDir, "%s/%s", baseDir, vm->def->name) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Failed to allocate memory for "
                       "snapshot directory for domain %s"),
                       vm->def->name);
        goto cleanup;
    }

    VIR_INFO("Scanning for snapshots for domain %s in %s", vm->def->name,
             snapDir);

    if (virDirOpenIfExists(&dir, snapDir) <= 0)
        goto cleanup;

    while ((direrr = virDirRead(dir, &entry, NULL)) > 0) {
        /* NB: ignoring errors, so one malformed config doesn't
           kill the whole process */
        VIR_INFO("Loading snapshot file '%s'", entry->d_name);

        if (virAsprintf(&fullpath, "%s/%s", snapDir, entry->d_name) < 0) {
            virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                           _("Failed to allocate memory for path"));
            continue;
        }

        if (virFileReadAll(fullpath, 1024*1024*1, &xmlStr) < 0) {
            /* Nothing we can do here, skip this one */
            virReportSystemError(errno,
                                 _("Failed to read snapshot file %s"),
                                 fullpath);
            VIR_FREE(fullpath);
            continue;
        }

        def = virDomainSnapshotDefParseString(xmlStr, caps,
                                              vzct_driver->xmlopt,
                                              flags);
        if (def == NULL) {
            /* Nothing we can do here, skip this one */
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Failed to parse snapshot XML from file '%s'"),
                           fullpath);
            VIR_FREE(fullpath);
            VIR_FREE(xmlStr);
            continue;
        }

        snap = virDomainSnapshotAssignDef(vm->snapshots, def);
        if (snap == NULL) {
            virDomainSnapshotDefFree(def);
        } else if (snap->def->current) {
            current = snap;
            if (!vm->current_snapshot)
                vm->current_snapshot = snap;
        }

        VIR_FREE(fullpath);
        VIR_FREE(xmlStr);
    }
    if (direrr < 0)
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Failed to fully read directory %s"),
                       snapDir);

    if (vm->current_snapshot != current) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Too many snapshots claiming to be current for domain %s"),
                       vm->def->name);
        vm->current_snapshot = NULL;
    }

    if (virDomainSnapshotUpdateRelations(vm->snapshots) < 0)
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Snapshots have inconsistent relations for domain %s"),
                       vm->def->name);

    /* FIXME: vzct keeps internal track of snapshots.  We can get access
     * to this info via the "info snapshots" monitor command for running
     * domains, or via "vzct-img snapshot -l" for shutoff domains.  It would
     * be nice to update our internal state based on that, but there is a
     * a problem.  vzct doesn't track all of the same metadata that we do.
     * In particular we wouldn't be able to fill in the <parent>, which is
     * pretty important in our metadata.
     */

    virResetLastError();

    ret = 0;
 cleanup:
    VIR_DIR_CLOSE(dir);
    VIR_FREE(snapDir);
    virObjectUnref(caps);
    virObjectUnlock(vm);
    return ret;
}

static char *
vzctDomainManagedSavePath(vzctDriverPtr driver, virDomainObjPtr vm)
{
    char *path = NULL;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);

    if (virAsprintf(&path, "%s/%s.save", cfg->saveDir, vm->def->name) < 0)
        goto cleanup;

 cleanup:
    virObjectUnref(cfg);
    return path;
}

static int
vzctDomainManagedSaveLoad(virDomainObjPtr vm,
                          void *opaque)
{
    vzctDriverPtr driver = opaque;
    char *dumppath = NULL;
    int ret = -1;

    virObjectLock(vm);

    if (!(dumppath = vzctDomainManagedSavePath(driver, vm)))
        goto cleanup;

    vm->hasManagedSave = virFileExists(dumppath);

    ret = 0;
 cleanup:
    virObjectUnlock(vm);
    VIR_FREE(dumppath);
    return ret;
}

static int
vzctDomainFindMaxID(virDomainObjPtr vm, void *data)
{
    int *maxid = data;

    if (vm->def->id > *maxid)
        *maxid = vm->def->id;

    return 0;
}

static void
vzctMonitorEvent(int watch ATTRIBUTE_UNUSED,
                 int fd,
                 int events,
                 void *opaque)
{
    vzctDriverPtr driver = opaque;
    struct sockaddr_nl addr;
    struct msghdr msg = { .msg_control = 0 };
    struct iovec iov;
    char buf[512];
    char *token, *saveptr;
    int len;
    virDomainObjPtr vm;
    unsigned char uuid[VIR_UUID_BUFLEN];
    vzctDomainJobPtr job;

    if ((events & (VIR_EVENT_HANDLE_HANGUP | VIR_EVENT_HANDLE_ERROR))) {
        VIR_WARN("Unexpected event socket event %d", events);
        return;
    }

    msg.msg_name = (void*)&addr;
    msg.msg_namelen = sizeof(addr);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    iov.iov_base = buf;
    iov.iov_len = sizeof(buf) - 1;

    if ((len = recvmsg(fd, &msg, MSG_DONTWAIT)) < 0) {
        VIR_ERROR(_("Can not read event message, error: %d"), errno);
        return;
    }

    buf[len] = '\0';

    if (!(token = strtok_r(buf, "@", &saveptr)) || STRNEQ(token, "ve-stop"))
        return;

    if (!(token = strtok_r(NULL, "@", &saveptr))) {
        VIR_ERROR(_("Can not read stopped container uuid"));
        return;
    }

    if (virUUIDParse(token, uuid) < 0) {
        VIR_ERROR(_("Can not parse stopped container uuid: %s"), token);
        return;
    }

    if (!(vm = virDomainObjListFindByUUIDRef(driver->domains, uuid))) {
        VIR_INFO("Can not find stopped container bu uuid: %s", token);
        return;
    }

    if (VIR_ALLOC(job) < 0)
        goto cleanup;

    job->type = VZCT_DOMAIN_JOB_SHUTDOWN;
    job->vm = vm;

    virObjectRef(vm);
    if (virThreadPoolSendJob(driver->workerPool, 0, job) < 0) {
        virObjectUnref(vm);
        VIR_FREE(job);
        goto cleanup;
    }

 cleanup:
    virDomainObjEndAPI(&vm);
}

static int
vzctStartMonitor(vzctDriverPtr driver)
{
    struct sockaddr_nl addr = { 0 };

    if ((driver->eventfd = socket(PF_NETLINK, SOCK_DGRAM, 31)) < 0) {
        virReportSystemError(errno,
                             "%s", _("failed to create event socket"));
        return -1;
    }

    if (virSetNonBlock(driver->eventfd) < 0) {
        virReportSystemError(errno, "%s",
                             _("Unable to put event socket into non-blocking mode"));
        return -1;
    }

    if (virSetCloseExec(driver->eventfd) < 0) {
        virReportSystemError(errno, "%s",
                             _("Unable to set event socket close-on-exec flag"));
        return -1;
    }

    addr.nl_family = AF_NETLINK;
    addr.nl_groups = 0x01; /* multicast */

    if (bind(driver->eventfd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        virReportSystemError(errno, "%s",
                             _("Unable to bind event socket"));
        return -1;
    }

    if ((driver->eventmon = virEventAddHandle(driver->eventfd,
                                              VIR_EVENT_HANDLE_HANGUP |
                                              VIR_EVENT_HANDLE_ERROR |
                                              VIR_EVENT_HANDLE_READABLE,
                                              vzctMonitorEvent,
                                              driver,
                                              NULL)) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("unable to add event socket to io loop"));
        return -1;
    }

    return 0;
}

static void
vzctStopMonitor(vzctDriverPtr driver)
{
    if (driver->eventmon)
        virEventRemoveHandle(driver->eventmon);
    VIR_FORCE_CLOSE(driver->eventfd);
}

static void
vzctHanledShutdown(vzctDriverPtr driver, virDomainObjPtr vm)
{
    bool job;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        return;

    if (!virDomainObjIsActive(vm))
        goto cleanup;

    vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_SHUTDOWN);

 cleanup:
    vzctDomainObjEndJob(vm, &job);
}

static void
vzctDomainJobHandler(void *data, void *opaque)
{
    vzctDomainJobPtr job = data;
    virDomainObjPtr vm = job->vm;
    vzctDriverPtr driver = opaque;

    virObjectLock(vm);
    switch (job->type) {
    case VZCT_DOMAIN_JOB_SHUTDOWN:
        vzctHanledShutdown(driver, vm);
        break;
    case VZCT_DOMAIN_JOB_LAST:
        break;
    }

    virDomainObjEndAPI(&vm);
    VIR_FREE(job);
}

static int
vzctStateInitialize(bool privileged ATTRIBUTE_UNUSED,
                    virStateInhibitCallback callback ATTRIBUTE_UNUSED,
                    void *opaque ATTRIBUTE_UNUSED)
{
    vzctDriverConfigPtr cfg = NULL;

    /* Check that the user is root, silently disable if not */
    if (!privileged) {
        VIR_INFO("Not running privileged, disabling driver");
        return 0;
    }

    if (VIR_ALLOC(vzct_driver) < 0)
        return -1;

    if (virMutexInit(&vzct_driver->lock) < 0) {
        VIR_FREE(vzct_driver);
        return -1;
    }

    if (!(vzct_driver->domains = virDomainObjListNew()))
        goto cleanup;

    vzct_driver->domainEventState = virObjectEventStateNew();
    if (!vzct_driver->domainEventState)
        goto cleanup;

    vzct_driver->hostsysinfo = virSysinfoRead();


    if (!(vzct_driver->config = cfg = vzctDriverConfigNew()))
        goto cleanup;

    if (vzctLoadDriverConfig(cfg, SYSCONFDIR "/libvirt/vzct.conf") < 0)
        goto cleanup;

    if (!(vzct_driver->caps = vzctDriverCapsInit()))
        goto cleanup;

    if (!(vzct_driver->xmlopt = vzctDomainXMLConfInit()))
        goto cleanup;

    if (!(vzct_driver->closeCallbacks = virCloseCallbacksNew()))
        goto cleanup;

    if (!(vzct_driver->vncPorts =
          virPortAllocatorRangeNew(_("vnc"), 5900, 65535)))
        goto cleanup;

    if (!(vzct_driver->migrationPorts =
          virPortAllocatorRangeNew(_("migration"), 49400, 49600)))
        goto cleanup;

    if (virFileMakePath(cfg->stateDir) < 0) {
        virReportSystemError(errno,
                             _("Failed to mkdir %s"),
                             cfg->stateDir);
        goto cleanup;
    }

    if (virFileMakePath(cfg->saveDir) < 0) {
        virReportSystemError(errno, _("Failed to mkdir %s"), cfg->saveDir);
        goto cleanup;
    }

    if (vzctStartMonitor(vzct_driver) < 0)
        goto cleanup;

    /* first load status files */
    if (virDomainObjListLoadAllConfigs(vzct_driver->domains,
                                       cfg->stateDir,
                                       cfg->autostartDir, 1,
                                       vzct_driver->caps,
                                       vzct_driver->xmlopt,
                                       NULL, NULL) < 0)
        goto cleanup;

    virDomainObjListForEach(vzct_driver->domains,
                            vzctDomainFindMaxID,
                            &vzct_driver->lastvmid);

    /* second load config files */
    if (virDomainObjListLoadAllConfigs(vzct_driver->domains,
                                       cfg->configDir,
                                       "/etc/libvirt/vzct-autostart", 0,
                                       vzct_driver->caps,
                                       vzct_driver->xmlopt,
                                       NULL, NULL) < 0)
        goto cleanup;

    virDomainObjListForEachUnlocked(vzct_driver->domains,
                                    vzctCheckState,
                                    vzct_driver);

    virDomainObjListForEach(vzct_driver->domains,
                            vzctDomainSnapshotLoad,
                            cfg->snapshotDir);

    virDomainObjListForEach(vzct_driver->domains,
                            vzctDomainManagedSaveLoad,
                            vzct_driver);

    vzct_driver->workerPool = virThreadPoolNew(0, 1, 0,
                                               vzctDomainJobHandler,
                                               vzct_driver);

    return 0;

 cleanup:
    vzctStateCleanup();
    return -1;
}

static int
vzctConnectGetVersion(virConnectPtr conn, unsigned long *hvVer)
{
    if (virConnectGetVersionEnsureACL(conn) < 0)
        return -1;

    *hvVer = 111;
    return 0;
}

static char *
vzctConnectGetHostname(virConnectPtr conn)
{
    if (virConnectGetHostnameEnsureACL(conn) < 0)
        return NULL;

    return virGetHostname();
}

static char *
vzctConnectGetSysinfo(virConnectPtr conn, unsigned int flags)
{
    vzctDriverPtr driver = conn->privateData;
    virBuffer buf = VIR_BUFFER_INITIALIZER;

    virCheckFlags(0, NULL);

    if (virConnectGetSysinfoEnsureACL(conn) < 0)
        return NULL;

    if (!driver->hostsysinfo) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Host SMBIOS information is not available"));
        return NULL;
    }

    if (virSysinfoFormat(&buf, driver->hostsysinfo) < 0)
        return NULL;
    if (virBufferCheckError(&buf) < 0)
        return NULL;

    return virBufferContentAndReset(&buf);
}

static int
vzctConnectGetMaxVcpus(virConnectPtr conn, const char *type)
{
    if (virConnectGetMaxVcpusEnsureACL(conn) < 0)
        return -1;

    /* As far as we have no limitation for containers
     * we report maximum */
    if (type == NULL || STRCASEEQ(type, "vzct"))
        return 1028;

    virReportError(VIR_ERR_INVALID_ARG,
                   _("unknown type '%s'"), type);
    return -1;
}

static int
vzctNodeGetInfo(virConnectPtr conn, virNodeInfoPtr nodeinfo)
{
    if (virNodeGetInfoEnsureACL(conn) < 0)
        return -1;

    return virCapabilitiesGetNodeInfo(nodeinfo);
}

static int
vzctNodeGetCPUStats(virConnectPtr conn,
                    int cpuNum,
                    virNodeCPUStatsPtr params,
                    int *nparams,
                    unsigned int flags)
{
    if (virNodeGetCPUStatsEnsureACL(conn) < 0)
        return -1;

    return virHostCPUGetStats(cpuNum, params, nparams, flags);
}

static int
vzctNodeGetMemoryStats(virConnectPtr conn,
                       int cellNum,
                       virNodeMemoryStatsPtr params,
                       int *nparams,
                       unsigned int flags)
{
    if (virNodeGetMemoryStatsEnsureACL(conn) < 0)
        return -1;

    return virHostMemGetStats(cellNum, params, nparams, flags);
}

static int
vzctNodeGetCellsFreeMemory(virConnectPtr conn,
                           unsigned long long *freeMems,
                           int startCell,
                           int maxCells)
{
    if (virNodeGetCellsFreeMemoryEnsureACL(conn) < 0)
        return -1;

    return virHostMemGetCellsFree(freeMems, startCell, maxCells);
}

static unsigned long long
vzctNodeGetFreeMemory(virConnectPtr conn)
{
    unsigned long long freeMem;

    if (virNodeGetFreeMemoryEnsureACL(conn) < 0)
        return -1;

    if (virHostMemGetInfo(NULL, &freeMem) < 0)
        return 0;
    return freeMem;
}

static int
vzctNodeGetCPUMap(virConnectPtr conn,
                  unsigned char **cpumap,
                  unsigned int *online,
                  unsigned int flags)
{
    if (virNodeGetCPUMapEnsureACL(conn) < 0)
        return -1;

    return virHostCPUGetMap(cpumap, online, flags);
}

static char *
vzctConnectGetCapabilities(virConnectPtr conn)
{
    vzctDriverPtr driver = conn->privateData;
    char *xml;

    if (virConnectGetCapabilitiesEnsureACL(conn) < 0)
        return NULL;

    xml = virCapabilitiesFormatXML(driver->caps);
    return xml;
}

static int
vzctConnectIsEncrypted(virConnectPtr conn ATTRIBUTE_UNUSED)
{
    /* Encryption is not relevant / applicable to way we talk to PCS */
    return 0;
}

static int
vzctConnectIsSecure(virConnectPtr conn ATTRIBUTE_UNUSED)
{
    /* We run CLI tools directly so this is secure */
    return 1;
}

static int
vzctConnectIsAlive(virConnectPtr conn ATTRIBUTE_UNUSED)
{
    return 1;
}

static virDomainPtr
vzctDomainDefineXMLFlags(virConnectPtr conn,
                         const char *xml,
                         unsigned int flags)
{
    vzctDriverPtr driver = conn->privateData;
    virDomainDefPtr def = NULL;
    virDomainObjPtr vm = NULL;
    virDomainPtr dom = NULL;
    virObjectEventPtr event = NULL;
    virDomainDefPtr oldDef = NULL;
    virCapsPtr caps = NULL;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE;

    virCheckFlags(VIR_DOMAIN_DEFINE_VALIDATE, NULL);

    if (flags & VIR_DOMAIN_DEFINE_VALIDATE)
        parse_flags |= VIR_DOMAIN_DEF_PARSE_VALIDATE_SCHEMA;

    caps = vzctDriverGetCapabilities(driver);

    if (!(def = virDomainDefParseString(xml, caps, driver->xmlopt,
                                        NULL, parse_flags)))
        goto cleanup;

    if (virXMLCheckIllegalChars("name", def->name, "\n") < 0)
        goto cleanup;

    if (virDomainDefineXMLFlagsEnsureACL(conn, def) < 0)
        goto cleanup;

    if (!(vm = virDomainObjListAdd(driver->domains, def,
                                   driver->xmlopt,
                                   0, &oldDef)))
        goto cleanup;

    virObjectRef(vm);

    def = NULL;
    vm->persistent = 1;

    if (virDomainSaveConfig(cfg->configDir, driver->caps,
                            vm->newDef ? vm->newDef : vm->def) < 0) {
        if (oldDef) {
            VIR_INFO("Restoring domain '%s' definition", vm->def->name);
            virDomainObjAssignDef(vm, oldDef, false, NULL);
            oldDef = NULL;
        } else {
            /* Brand new domain. Remove it */
            VIR_INFO("Deleting domain '%s'", vm->def->name);
            virDomainObjListRemove(driver->domains, vm);
        }
        goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(vm,
                                     VIR_DOMAIN_EVENT_DEFINED,
                                     !oldDef ?
                                     VIR_DOMAIN_EVENT_DEFINED_ADDED :
                                     VIR_DOMAIN_EVENT_DEFINED_UPDATED);

    dom = virGetDomain(conn, vm->def->name, vm->def->uuid, vm->def->id);

 cleanup:
    virDomainDefFree(def);
    virDomainDefFree(oldDef);
    virDomainObjEndAPI(&vm);
    if (event)
        virObjectEventStateQueue(driver->domainEventState, event);
    virObjectUnref(caps);
    virObjectUnref(cfg);
    return dom;
}

static virDomainPtr
vzctDomainDefineXML(virConnectPtr conn, const char *xml)
{
    return vzctDomainDefineXMLFlags(conn, xml, 0);
}

static char
*vzctDomainGetXMLDesc(virDomainPtr dom,
                      unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm;
    char *ret = NULL;

    /* Flags checked by virDomainDefFormat */

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainGetXMLDescEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    ret = virDomainDefFormat((flags & VIR_DOMAIN_XML_INACTIVE) &&
                             vm->newDef ? vm->newDef : vm->def,
                             driver->caps,
                             virDomainDefFormatConvertXMLFlags(flags));

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainUndefineFlags(virDomainPtr dom, unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm;
    virObjectEventPtr event = NULL;
    int ret = -1;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    char *dumppath = NULL;

    virCheckFlags(VIR_DOMAIN_UNDEFINE_MANAGED_SAVE, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainUndefineFlagsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (!vm->persistent) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Cannot undefine transient domain"));
        goto cleanup;
    }

    if (!(dumppath = vzctDomainManagedSavePath(driver, vm)))
        goto cleanup;

    if (virFileExists(dumppath)) {
        if (flags & VIR_DOMAIN_UNDEFINE_MANAGED_SAVE) {
            if (virFileDeleteTree(dumppath) < 0) {
                virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                               _("Failed to remove domain managed save state"));
                goto cleanup;
            }
        } else {
            virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                           _("Refusing to undefine while domain managed "
                             "save image exists"));
            goto cleanup;
        }
    }

    if (virDomainDeleteConfig(cfg->configDir,
                              cfg->autostartDir,
                              vm) < 0)
        goto cleanup;

    event = virDomainEventLifecycleNewFromObj(vm,
                                     VIR_DOMAIN_EVENT_UNDEFINED,
                                     VIR_DOMAIN_EVENT_UNDEFINED_REMOVED);

    if (virDomainObjIsActive(vm))
        vm->persistent = 0;
    else
        virDomainObjListRemove(driver->domains, vm);

    ret = 0;

 cleanup:
    virDomainObjEndAPI(&vm);
    if (event)
        virObjectEventStateQueue(driver->domainEventState, event);
    virObjectUnref(cfg);
    VIR_FREE(dumppath);
    return ret;
}

static int
vzctDomainUndefine(virDomainPtr dom)
{
    return vzctDomainUndefineFlags(dom, 0);
}

static virDomainPtr
vzctDomainLookupByID(virConnectPtr conn, int id)
{
    vzctDriverPtr driver = conn->privateData;
    virDomainObjPtr vm;
    virDomainPtr dom = NULL;

    vm = virDomainObjListFindByID(driver->domains, id);

    if (!vm) {
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("No domain with matching id %d"), id);
        goto cleanup;
    }

    if (virDomainLookupByIDEnsureACL(conn, vm->def) < 0)
        goto cleanup;

    dom = virGetDomain(conn, vm->def->name, vm->def->uuid, vm->def->id);

 cleanup:
    if (vm)
        virObjectUnlock(vm);
    return dom;
}

static virDomainPtr
vzctDomainLookupByUUID(virConnectPtr conn, const unsigned char *uuid)
{
    vzctDriverPtr driver = conn->privateData;
    virDomainObjPtr vm;
    virDomainPtr dom = NULL;

    vm = virDomainObjListFindByUUIDRef(driver->domains, uuid);

    if (!vm) {
        char uuidstr[VIR_UUID_STRING_BUFLEN];
        virUUIDFormat(uuid, uuidstr);
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("No domain with matching uuid '%s'"), uuidstr);
        goto cleanup;
    }

    if (virDomainLookupByUUIDEnsureACL(conn, vm->def) < 0)
        goto cleanup;

    dom = virGetDomain(conn, vm->def->name, vm->def->uuid, vm->def->id);

 cleanup:
    virDomainObjEndAPI(&vm);
    return dom;
}

static virDomainPtr
vzctDomainLookupByName(virConnectPtr conn, const char *name)
{
    vzctDriverPtr driver = conn->privateData;
    virDomainObjPtr vm;
    virDomainPtr dom = NULL;

    vm = virDomainObjListFindByName(driver->domains, name);
    if (!vm) {
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("No domain with matching name '%s'"), name);
        goto cleanup;
    }

    if (virDomainLookupByNameEnsureACL(conn, vm->def) < 0)
        goto cleanup;

    dom = virGetDomain(conn, vm->def->name, vm->def->uuid, vm->def->id);

 cleanup:
    virDomainObjEndAPI(&vm);
    return dom;
}


static int
vzctDomainIsActive(virDomainPtr dom)
{
    virDomainObjPtr obj;
    int ret = -1;

    if (!(obj = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainIsActiveEnsureACL(dom->conn, obj->def) < 0)
        goto cleanup;

    ret = virDomainObjIsActive(obj);

 cleanup:
    virDomainObjEndAPI(&obj);
    return ret;
}

static int
vzctDomainIsPersistent(virDomainPtr dom)
{
    virDomainObjPtr obj;
    int ret = -1;

    if (!(obj = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainIsPersistentEnsureACL(dom->conn, obj->def) < 0)
        goto cleanup;

    ret = obj->persistent;

 cleanup:
    virDomainObjEndAPI(&obj);
    return ret;
}

static int
vzctDomainIsUpdated(virDomainPtr dom)
{
    virDomainObjPtr obj;
    int ret = -1;

    if (!(obj = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainIsUpdatedEnsureACL(dom->conn, obj->def) < 0)
        goto cleanup;

    ret = obj->updated;

 cleanup:
    virDomainObjEndAPI(&obj);
    return ret;
}

static int
vzctConnectListDomains(virConnectPtr conn, int *ids, int nids)
{
    vzctDriverPtr driver = conn->privateData;
    int n;

    if (virConnectListDomainsEnsureACL(conn) < 0)
        return -1;

    n = virDomainObjListGetActiveIDs(driver->domains, ids, nids,
                                     virConnectListDomainsCheckACL, conn);

    return n;
}

static int
vzctConnectNumOfDomains(virConnectPtr conn)
{
    vzctDriverPtr driver = conn->privateData;
    int n;

    if (virConnectNumOfDomainsEnsureACL(conn) < 0)
        return -1;

    n = virDomainObjListNumOfDomains(driver->domains, true,
                                     virConnectNumOfDomainsCheckACL, conn);

    return n;
}

static int
vzctConnectListDefinedDomains(virConnectPtr conn,
                              char **const names,
                              int nnames)
{
    vzctDriverPtr driver = conn->privateData;
    int n;

    if (virConnectListDefinedDomainsEnsureACL(conn) < 0)
        return -1;

    n = virDomainObjListGetInactiveNames(driver->domains, names, nnames,
                                         virConnectListDefinedDomainsCheckACL, conn);

    return n;
}

static int
vzctConnectNumOfDefinedDomains(virConnectPtr conn)
{
    vzctDriverPtr driver = conn->privateData;
    int n;

    if (virConnectNumOfDefinedDomainsEnsureACL(conn) < 0)
        return -1;

    n = virDomainObjListNumOfDomains(driver->domains, false,
                                     virConnectNumOfDefinedDomainsCheckACL, conn);

    return n;
}

static int
vzctConnectListAllDomains(virConnectPtr conn,
                          virDomainPtr **domains,
                          unsigned int flags)
{
    vzctDriverPtr driver = conn->privateData;
    int ret = -1;

    virCheckFlags(VIR_CONNECT_LIST_DOMAINS_FILTERS_ALL, -1);

    if (virConnectListAllDomainsEnsureACL(conn) < 0)
        return -1;

    ret = virDomainObjListExport(driver->domains, conn, domains,
                                 virConnectListAllDomainsCheckACL, flags);
    return ret;
}

static int
vzctDomainGetInfo(virDomainPtr dom, virDomainInfoPtr info)
{
    virDomainObjPtr vm;
    int ret = -1;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainGetInfoEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    info->state = virDomainObjGetState(vm, NULL);

    /* TODO */
    info->cpuTime = 0;
    info->memory = vm->def->mem.cur_balloon;

    info->maxMem = virDomainDefGetMemoryTotal(vm->def);
    info->nrVirtCpu = virDomainDefGetVcpus(vm->def);
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainGetState(virDomainPtr dom,
                   int *state,
                   int *reason,
                   unsigned int flags)
{
    virDomainObjPtr vm;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainGetStateEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    *state = virDomainObjGetState(vm, reason);
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctStartVNCServer(vzctDriverPtr driver,
                   virDomainObjPtr vm,
                   virDomainGraphicsDefPtr gr)
{
    vzctDomainObjPrivatePtr priv = vm->privateData;
    virCommandPtr cmd;
    char *pidfile = NULL;
    int ret = -1;

    if (!gr->listens[0].address &&
        VIR_STRDUP(gr->listens[0].address, "127.0.0.1") < 0)
        return -1;

    if (gr->data.vnc.autoport) {
        unsigned short port;

        if (virPortAllocatorAcquire(driver->vncPorts, &port) < 0)
            return -1;

        gr->data.vnc.port = port;
    } else if (virPortAllocatorSetUsed(gr->data.vnc.port) < 0) {
        return -1;
    }

    cmd = virCommandNew("/usr/bin/prl_vzvncserver_app");

    virCommandAddArgList(cmd, "--listen", gr->listens[0].address, NULL);
    virCommandAddArgFormat(cmd, "--port=%d", gr->data.vnc.port);

    if (gr->data.vnc.auth.passwd) {
        virCommandAddArg(cmd, "--passwd");
        virCommandSetInputBuffer(cmd, gr->data.vnc.auth.passwd);
    }

    virCommandAddArg(cmd, vm->def->name);

    if (virAsprintf(&pidfile, "/vz/vzct/%s.vnc.pid", vm->def->name) < 0)
        goto cleanup;

    if (unlink(pidfile) < 0 && errno != ENOENT) {
        virReportSystemError(errno, _("Cannot remove stale PID file %s"), pidfile);
        goto cleanup;
    }

    virCommandSetPidFile(cmd, pidfile);
    virCommandDaemonize(cmd);

    if (virCommandRun(cmd, NULL))
        goto cleanup;

    if (virPidFileReadPath(pidfile, &priv->vncPid) < 0) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Failed to read vnc server pid"));
        goto cleanup;
    }

    ret = 0;

 cleanup:
    if (priv->vncPid == -1)
        virPortAllocatorRelease(gr->data.vnc.port);

    virCommandFree(cmd);
    VIR_FREE(pidfile);

    return ret;
}

static void
vzctStopVnc(vzctDriverPtr driver ATTRIBUTE_UNUSED, virDomainObjPtr vm)
{
    vzctDomainObjPrivatePtr priv = vm->privateData;

    if (vm->def->ngraphics > 0 && priv->vncPid > 0) {
        virDomainGraphicsDefPtr gr = vm->def->graphics[0];

        virProcessKill(priv->vncPid, SIGTERM);
        priv->vncPid = -1;
        virPortAllocatorRelease(gr->data.vnc.port);
    }
}

static void
vzctCleanupRuntimeFiles(vzctDriverPtr driver, virDomainObjPtr vm)
{
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    char *path;
    char ebuf[1024];

    if (virAsprintf(&path, "%s/%s.xml", cfg->stateDir, vm->def->name) < 0)
        goto cleanup;

    if (unlink(path) < 0 && errno != ENOENT)
        VIR_WARN("Failed to remove status file for %s: %s",
                 vm->def->name, virStrerror(errno, ebuf, sizeof(ebuf)));
    VIR_FREE(path);

    if (virAsprintf(&path, "%s/%s.vnc.pid", cfg->stateDir, vm->def->name) < 0)
        goto cleanup;

    if (unlink(path) < 0 && errno != ENOENT) {
        VIR_WARN("Failed to remove vnc pid file for %s: %s",
                 vm->def->name, virStrerror(errno, ebuf, sizeof(ebuf)));
        goto cleanup;
    }
    VIR_FREE(path);

 cleanup:
    virObjectUnref(cfg);
}

static void
vzctDomainStop(vzctDriverPtr driver,
               virDomainObjPtr vm,
               virDomainShutoffReason reason)
{
    int detail = VIR_DOMAIN_EVENT_STOPPED_SHUTDOWN;
    const char *audit = NULL;

    /* don't check if domain is active here as some paths want to use
     * this function to remove transient definition */

    vm->def->id = -1;
    vzctStopVnc(driver, vm);
    vzctlDomainDestroy(vm, M_KILL_FORCE);
    virNetDevExeCleanupInterfaces(vm);
    virDomainObjRemoveTransientDef(vm);
    virDomainObjSetState(vm, VIR_DOMAIN_SHUTOFF, reason);

    switch (reason) {
    case VIR_DOMAIN_SHUTOFF_DESTROYED:
        detail = VIR_DOMAIN_EVENT_STOPPED_DESTROYED;
        audit = "destroyed";
        break;

    case VIR_DOMAIN_SHUTOFF_SHUTDOWN:
        detail = VIR_DOMAIN_EVENT_STOPPED_SHUTDOWN;
        audit = "shutdown";
        break;

    case VIR_DOMAIN_SHUTOFF_MIGRATED:
        detail = VIR_DOMAIN_EVENT_STOPPED_MIGRATED;
        audit = "migrated";
        break;

    case VIR_DOMAIN_SHUTOFF_SAVED:
        detail = VIR_DOMAIN_EVENT_STOPPED_SAVED;
        audit = "saved";
        break;

    case VIR_DOMAIN_SHUTOFF_FAILED:
        detail = VIR_DOMAIN_EVENT_STOPPED_FAILED;
        audit = "failed";
        break;

    case VIR_DOMAIN_SHUTOFF_UNKNOWN:
    case VIR_DOMAIN_SHUTOFF_CRASHED:
    case VIR_DOMAIN_SHUTOFF_FROM_SNAPSHOT:
    case VIR_DOMAIN_SHUTOFF_LAST:
        audit = NULL;
    }

    if (audit) {
        virObjectEventPtr event = NULL;

        event = virDomainEventLifecycleNewFromObj(vm,
                                                  VIR_DOMAIN_EVENT_STOPPED,
                                                  detail);
        if (event)
            virObjectEventStateQueue(driver->domainEventState, event);

        virDomainAuditStop(vm, audit);
    }

    if (!vm->persistent)
        virDomainObjListRemove(driver->domains, vm);

    vzctCleanupRuntimeFiles(driver, vm);
}

static int
vzctDomainObjStart(vzctDriverPtr driver,
                   virDomainObjPtr vm,
                   char *dumppath,
                   bool paused)
{
    int ret = -1;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);

    vm->def->id = virAtomicIntInc(&driver->lastvmid);

    if (virDomainObjSetDefTransient(driver->caps, driver->xmlopt, vm) < 0)
        goto cleanup;

    if (virNetDevExeSetupInterfaces(vm->def, cfg->stateDir) < 0)
        goto cleanup;

    if (vzctlDomainCreate(vm, paused, dumppath) < 0)
        goto cleanup;

    if (vm->def->ngraphics > 0 &&
        vzctStartVNCServer(driver, vm, vm->def->graphics[0]) < 0)
        goto cleanup;

    ret = 0;
 cleanup:

    virObjectUnref(cfg);
    return ret;
}

static int
vzctDomainObjBoot(vzctDriverPtr driver, virDomainObjPtr vm, bool paused)
{
    int ret = -1;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    virObjectEventPtr event;
    int reason, detail;
    char *dumppath = NULL;

    if (vm->hasManagedSave) {
        detail = VIR_DOMAIN_EVENT_STARTED_RESTORED;
        reason = VIR_DOMAIN_RUNNING_RESTORED;
        if (!(dumppath = vzctDomainManagedSavePath(driver, vm)))
            goto cleanup;
    } else {
        detail = VIR_DOMAIN_EVENT_STARTED_BOOTED;
        reason = VIR_DOMAIN_RUNNING_BOOTED;
    }

    if (vzctDomainObjStart(driver, vm, dumppath, paused) < 0)
        goto cleanup;

    if (paused)
        virDomainObjSetState(vm, VIR_DOMAIN_PAUSED, VIR_DOMAIN_PAUSED_USER);
    else
        virDomainObjSetState(vm, VIR_DOMAIN_RUNNING, reason);

    if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm, driver->caps) < 0)
        goto cleanup;

    if (vm->hasManagedSave) {
        if (virFileDeleteTree(dumppath) < 0)
            VIR_WARN("Failed to remove the managed state %s", dumppath);
        else
            vm->hasManagedSave = false;
    }

    event = virDomainEventLifecycleNewFromObj(vm,
                                              VIR_DOMAIN_EVENT_STARTED, detail);
    if (event)
        virObjectEventStateQueue(driver->domainEventState, event);

    if (paused) {
        event = virDomainEventLifecycleNewFromObj(vm,
                                         VIR_DOMAIN_EVENT_SUSPENDED,
                                         VIR_DOMAIN_EVENT_SUSPENDED_PAUSED);
        if (event)
            virObjectEventStateQueue(driver->domainEventState, event);
    }

    ret = 0;

 cleanup:
    virDomainAuditStart(vm, "booted", ret == 0);

    if (ret < 0)
        vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_FAILED);

    VIR_FREE(dumppath);
    virObjectUnref(cfg);
    return ret;
}


static int
vzctDomainCreateWithFlags(virDomainPtr dom, unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm;
    int ret = -1;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_START_PAUSED, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainCreateWithFlagsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Domain is already running"));
        goto cleanup;
    }

    if (vm->hasManagedSave && (flags & VIR_DOMAIN_PAUSED)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("Paused start from managed save is not supported"));
        goto cleanup;
    }

    ret = vzctDomainObjBoot(driver, vm, !!(flags & VIR_DOMAIN_PAUSED));

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainCreate(virDomainPtr dom)
{
    return vzctDomainCreateWithFlags(dom, 0);
}

static int
vzctDomainSuspend(virDomainPtr dom)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm;
    virObjectEventPtr event = NULL;
    int ret = -1;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    bool job = false;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainSuspendEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Domain is not running"));
        goto cleanup;
    }

    if (virDomainObjGetState(vm, NULL) != VIR_DOMAIN_PAUSED) {

        if (vzctlDomainPause(vm) < 0)
            goto cleanup;

        virDomainObjSetState(vm, VIR_DOMAIN_PAUSED, VIR_DOMAIN_PAUSED_USER);

        event = virDomainEventLifecycleNewFromObj(vm,
                                         VIR_DOMAIN_EVENT_SUSPENDED,
                                         VIR_DOMAIN_EVENT_SUSPENDED_PAUSED);
    }

    if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm, driver->caps) < 0)
        goto cleanup;
    ret = 0;

 cleanup:
    if (event)
        virObjectEventStateQueue(driver->domainEventState, event);
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    virObjectUnref(cfg);
    return ret;
}

static int
vzctDomainResume(virDomainPtr dom)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm;
    virObjectEventPtr event = NULL;
    int ret = -1;
    int state;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    bool job = false;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainResumeEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Domain is not running"));
        goto cleanup;
    }

    state = virDomainObjGetState(vm, NULL);
    if (state == VIR_DOMAIN_RUNNING) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is already running"));
        goto cleanup;
    } else if (state == VIR_DOMAIN_PAUSED) {

        if (vzctlDomainUnpause(vm) < 0)
            goto cleanup;

        virDomainObjSetState(vm, VIR_DOMAIN_RUNNING, VIR_DOMAIN_RUNNING_UNPAUSED);

        event = virDomainEventLifecycleNewFromObj(vm,
                                         VIR_DOMAIN_EVENT_RESUMED,
                                         VIR_DOMAIN_EVENT_RESUMED_UNPAUSED);
    }

    if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm, driver->caps) < 0)
        goto cleanup;
    ret = 0;

 cleanup:
    if (event)
        virObjectEventStateQueue(driver->domainEventState, event);
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    virObjectUnref(cfg);
    return ret;
}

static int
vzctDomainDestroyFlags(virDomainPtr dom,
                       unsigned int flags)
{
    virDomainObjPtr vm;
    vzctDriverPtr driver = dom->conn->privateData;
    int ret = -1;
    bool job = false;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainDestroyFlagsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Domain is not running"));
        goto cleanup;
    }

    if (vzctlDomainDestroy(vm, M_KILL_FORCE) < 0)
        goto cleanup;

    vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_DESTROYED);
    ret = 0;

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainDestroy(virDomainPtr dom)
{
    return vzctDomainDestroyFlags(dom, 0);
}

static int
vzctConnectDomainEventRegisterAny(virConnectPtr conn,
                                  virDomainPtr dom,
                                  int eventID,
                                  virConnectDomainEventGenericCallback callback,
                                  void *opaque,
                                  virFreeCallback freecb)
{
    int ret = -1;
    vzctDriverPtr driver = conn->privateData;

    if (virConnectDomainEventRegisterAnyEnsureACL(conn) < 0)
        return -1;

    if (virDomainEventStateRegisterID(conn,
                                      driver->domainEventState,
                                      dom, eventID,
                                      callback, opaque, freecb, &ret) < 0)
        ret = -1;
    return ret;
}

static int
vzctConnectDomainEventDeregisterAny(virConnectPtr conn,
                                    int callbackID)
{
    vzctDriverPtr driver = conn->privateData;

    if (virConnectDomainEventDeregisterAnyEnsureACL(conn) < 0)
        return -1;

    if (virObjectEventStateDeregisterID(conn,
                                        driver->domainEventState,
                                        callbackID, true) < 0)
        return -1;

    return 0;
}

static int
vzctDomainBlockStatsImpl(virDomainObjPtr vm,
                         const char *path,
                         virDomainBlockStatsPtr stats)
{
    int idx;

    if (*path == '\0')
        return vzctlDomainBlockStatsAll(vm, stats);

    if ((idx = virDomainDiskIndexByName(vm->def, path, false)) < 0) {
        virReportError(VIR_ERR_INVALID_ARG, _("invalid path: %s"), path);
        return -1;
    }

    return vzctlDomainBlockStats(vm, vm->def->disks[idx]->src->path, stats);
}

static int
vzctDomainBlockStatsFlags(virDomainPtr dom,
                          const char *path,
                          virTypedParameterPtr params,
                          int *nparams,
                          unsigned int flags)
{
    virDomainObjPtr vm;
    virDomainBlockStatsStruct stats;
    int ret = -1;

    virCheckFlags(VIR_TYPED_PARAM_STRING_OKAY, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainBlockStatsFlagsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainBlockStatsImpl(vm, path, &stats) < 0)
        goto cleanup;

    if (*nparams == 0) {
        *nparams = 2;
        ret = 0;
        goto cleanup;
    }

    if (*nparams > 0 &&
        virTypedParameterAssign(&params[0],
                                VIR_DOMAIN_BLOCK_STATS_READ_BYTES,
                                VIR_TYPED_PARAM_LLONG, stats.rd_bytes) < 0)
        goto cleanup;

    if (*nparams > 1 &&
        virTypedParameterAssign(&params[1],
                                VIR_DOMAIN_BLOCK_STATS_WRITE_BYTES,
                                VIR_TYPED_PARAM_LLONG, stats.wr_bytes) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainBlockStats(virDomainPtr dom,
                     const char *path,
                     virDomainBlockStatsPtr stats)
{
    virDomainObjPtr vm;
    int ret = -1;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainBlockStatsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    ret = vzctDomainBlockStatsImpl(vm, path, stats);

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainMemoryStats(virDomainPtr dom,
                      virDomainMemoryStatPtr stats,
                      unsigned int nr_stats,
                      unsigned int flags)
{
    virDomainObjPtr vm;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainMemoryStatsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    ret = vzctlDomainMemoryStats(vm, stats, nr_stats);

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainInterfaceStats(virDomainPtr dom,
                         const char *path,
                         virDomainInterfaceStatsPtr stats)
{
    virDomainObjPtr vm;
    size_t i;
    int ret = -1;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainInterfaceStatsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    for (i = 0; i < vm->def->nnets; i++) {
        if (vm->def->nets[i]->ifname &&
            STREQ(vm->def->nets[i]->ifname, path)) {
            break;
        }
    }

    if (i == vm->def->nnets) {
        virReportError(VIR_ERR_INVALID_ARG,
                       _("invalid path, '%s' is not a known interface"), path);
        goto cleanup;
    }

    ret = virNetDevTapInterfaceStats(path, stats,
                                     virDomainNetTypeSharesHostView(vm->def->nets[i]));

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainObjAttachDevice(vzctDriverPtr driver,
                          virDomainObjPtr vm,
                          virDomainDefPtr def,
                          virDomainDeviceDefPtr dev,
                          bool live)
{
    switch ((virDomainDeviceType) dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        if (VIR_REALLOC_N(def->disks, def->ndisks + 1) < 0)
            return -1;

        if (live && vzctlAttachDisk(vm, dev) < 0)
            return -1;

        virDomainDiskInsertPreAlloced(def, dev->data.disk);
        dev->data.disk = NULL;
        break;

    case VIR_DOMAIN_DEVICE_FS:
        if (VIR_REALLOC_N(def->fss, def->nfss + 1) < 0)
            return -1;

        if (live && vzctlAttachDisk(vm, dev) < 0)
            return -1;

        ignore_value(VIR_INSERT_ELEMENT_INPLACE(def->fss, -1,
                                                def->nfss, dev->data.fs));
        dev->data.fs = NULL;
        break;

    case VIR_DOMAIN_DEVICE_NET:
        if (VIR_REALLOC_N(def->nets, def->nnets + 1) < 0)
            return -1;

        if (live && vzctlAttachNet(vm, dev->data.net))
            return -1;

        ignore_value(VIR_INSERT_ELEMENT_INPLACE(def->nets, -1,
                                                def->nnets, dev->data.net));
        dev->data.net = NULL;
        break;

    case VIR_DOMAIN_DEVICE_GRAPHICS:
        if (def->ngraphics > 0) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("Only one graphics per domain is supported."));
            return -1;
        }

        if (VIR_REALLOC_N(def->graphics, def->ngraphics + 1) < 0)
            return -1;

        if (live && vzctStartVNCServer(driver, vm, dev->data.graphics) < 0)
            return -1;

        ignore_value(VIR_INSERT_ELEMENT_INPLACE(def->graphics, -1,
                                                def->ngraphics, dev->data.graphics));
        dev->data.graphics = NULL;
        break;

    default:
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                       _("Attaching of device type '%s' is not supported"),
                       virDomainDeviceTypeToString(dev->type));
        return -1;
    }

    return 0;
}

static int
vzctDomainAttachDeviceFlags(virDomainPtr dom,
                            const char *xml,
                            unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm = NULL;
    int ret = -1;
    vzctDriverConfigPtr cfg = NULL;
    virDomainDeviceDefPtr dev_cfg = NULL, dev_live = NULL;
    virCapsPtr caps = NULL;
    virDomainDefPtr vmdef = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                               VIR_DOMAIN_DEF_PARSE_ABI_UPDATE;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainAttachDeviceFlagsEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (virDomainObjUpdateModificationImpact(vm, &flags) < 0)
        goto cleanup;

    cfg = vzctDriverGetConfig(driver);
    caps = vzctDriverGetCapabilities(driver);

    if (flags & VIR_DOMAIN_AFFECT_LIVE) {
        if (!(dev_live = virDomainDeviceDefParse(xml, vm->def,
                                           caps, driver->xmlopt,
                                           parse_flags)))
            goto cleanup;

        if (vzctDomainObjAttachDevice(driver, vm, vm->def, dev_live, true) < 0)
            goto cleanup;

        if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm,
                                driver->caps) < 0)
            goto cleanup;

        /* no rollback in case saving status was failed, by design */
    }

    if (flags & VIR_DOMAIN_AFFECT_CONFIG) {
        if (!(dev_cfg = virDomainDeviceDefParse(xml, vm->def,
                                                caps, driver->xmlopt,
                                                parse_flags)))
            goto cleanup;

        if (!(vmdef = virDomainObjCopyPersistentDef(vm, caps, driver->xmlopt)))
            goto cleanup;

        if (vzctDomainObjAttachDevice(driver, vm, vmdef, dev_cfg, false) < 0)
            goto cleanup;

        if (virDomainSaveConfig(cfg->configDir, driver->caps, vmdef) < 0)
            goto cleanup;

        virDomainObjAssignDef(vm, vmdef, false, NULL);
        vmdef = NULL;
    }

    ret = 0;

 cleanup:
    virDomainDefFree(vmdef);
    virDomainDeviceDefFree(dev_live);
    virDomainDeviceDefFree(dev_cfg);
    virObjectUnref(cfg);
    virObjectUnref(caps);
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);

    return ret;
}

static int
vzctDomainAttachDevice(virDomainPtr dom, const char *xml)
{
    return vzctDomainAttachDeviceFlags(dom, xml, VIR_DOMAIN_AFFECT_LIVE);
}

static int
vzctDomainNetFindIdx(virDomainDefPtr def, virDomainNetDefPtr net)
{
    size_t i;
    int matchidx = -1;

    for (i = 0; i < def->nnets; i++) {
        if (STREQ_NULLABLE(def->nets[i]->ifname_guest, net->ifname_guest)) {
            matchidx = i;
            break;
        }
    }

    if (matchidx < 0) {
        virReportError(VIR_ERR_OPERATION_FAILED,
                       _("no device matching guest name %s found."),
                       net->ifname_guest);
    }

    return matchidx;
}

static int
vzctDomainObjDetachDevice(vzctDriverPtr driver,
                          virDomainObjPtr vm,
                          virDomainDefPtr def,
                          virDomainDeviceDefPtr dev,
                          bool live)
{
    virDomainNetDefPtr net;
    virDomainFSDefPtr fs;
    virDomainDiskDefPtr disk;
    int idx;

    switch ((virDomainDeviceType) dev->type) {
    case VIR_DOMAIN_DEVICE_DISK:
        disk = dev->data.disk;
        idx = virDomainDiskIndexByName(def, disk->dst, false);
        if (idx < 0) {
            virReportError(VIR_ERR_OPERATION_FAILED, "%s",
                           _("no matching disk device was found"));
            return -1;
        }

        if (live && vzctlDetachDisk(vm, dev) < 0)
            return -1;

        virDomainDiskDefFree(virDomainDiskRemove(def, idx));
        break;

    case VIR_DOMAIN_DEVICE_FS:
        fs = dev->data.fs;
        idx = virDomainFSIndexByName(def, fs->dst);
        if (idx < 0) {
            virReportError(VIR_ERR_OPERATION_FAILED, "%s",
                           _("no matching filesystem device was found"));
            return -1;
        }

        if (live && vzctlDetachDisk(vm, dev) < 0)
            return -1;

        virDomainFSDefFree(virDomainFSRemove(def, idx));
        break;

    case VIR_DOMAIN_DEVICE_NET:
        net = dev->data.net;
        if ((idx = vzctDomainNetFindIdx(def, net)) < 0)
            return -1;

        if (live && vzctlDetachNet(vm, net) < 0)
            return -1;

        virDomainNetDefFree(virDomainNetRemove(def, idx));
        break;

    case VIR_DOMAIN_DEVICE_GRAPHICS:
        if (def->ngraphics == 0) {
            virReportError(VIR_ERR_INVALID_ARG, "%s",
                           _("Domain has no vnc graphics."));
            return -1;
        }

        if (live)
            vzctStopVnc(driver, vm);

        virDomainGraphicsDefFree(def->graphics[0]);
        VIR_DELETE_ELEMENT(def->graphics, 0, def->ngraphics);
        break;

    default:
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                       _("detaching device of type '%s' is not supported"),
                       virDomainDeviceTypeToString(dev->type));
        return -1;
    }

    return 0;
}

static int
vzctDomainDetachDeviceFlags(virDomainPtr dom,
                            const char *xml,
                            unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm = NULL;
    vzctDriverConfigPtr cfg = NULL;
    virDomainDeviceDefPtr dev = NULL;
    int ret = -1;
    virCapsPtr caps = NULL;
    virDomainDefPtr vmdef = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                               VIR_DOMAIN_DEF_PARSE_ABI_UPDATE;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainDetachDeviceFlagsEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (virDomainObjUpdateModificationImpact(vm, &flags) < 0)
        goto cleanup;

    if (!(dev = virDomainDeviceDefParse(xml, vm->def,
                                        caps, driver->xmlopt,
                                        parse_flags)))
        goto cleanup;

    cfg = vzctDriverGetConfig(driver);
    caps = vzctDriverGetCapabilities(driver);

    if (flags & VIR_DOMAIN_AFFECT_LIVE) {
        if (vzctDomainObjDetachDevice(driver, vm, vm->def, dev, true) < 0)
            goto cleanup;

        if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm,
                                driver->caps) < 0)
            goto cleanup;

        /* no rollback in case saving status was failed, by design */
    }

    if (flags & VIR_DOMAIN_AFFECT_CONFIG) {
        if (!(vmdef = virDomainObjCopyPersistentDef(vm, caps, driver->xmlopt)))
            goto cleanup;

        if (vzctDomainObjDetachDevice(driver, vm, vmdef, dev, false) < 0)
            goto cleanup;

        if (virDomainSaveConfig(cfg->configDir, driver->caps, vmdef) < 0)
            goto cleanup;

        virDomainObjAssignDef(vm, vmdef, false, NULL);
        vmdef = NULL;
    }

    ret = 0;

 cleanup:
    virDomainDefFree(vmdef);
    virDomainDeviceDefFree(dev);
    virObjectUnref(cfg);
    virObjectUnref(caps);
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainDetachDevice(virDomainPtr dom, const char *xml)
{
    return vzctDomainDetachDeviceFlags(dom, xml, VIR_DOMAIN_AFFECT_LIVE);
}

static int
vzctDomainObjUpdateDevice(vzctDriverPtr driver,
                          virDomainObjPtr vm,
                          virDomainDefPtr def,
                          virDomainDeviceDefPtr dev,
                          bool live)
{
    switch ((virDomainDeviceType) dev->type) {
    case VIR_DOMAIN_DEVICE_GRAPHICS:
        if (def->ngraphics == 0) {
            virReportError(VIR_ERR_INVALID_ARG, "%s",
                           _("Domain has no vnc graphics."));
            return -1;
        }

        if (live) {
            vzctStopVnc(driver, vm);
            if (vzctStartVNCServer(driver, vm, dev->data.graphics) < 0)
                return -1;
        }

        virDomainGraphicsDefFree(def->graphics[0]);
        def->graphics[0] = dev->data.graphics;
        dev->data.graphics = NULL;
        break;

    default:
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED,
                       _("Updating of device type '%s' is not supported"),
                       virDomainDeviceTypeToString(dev->type));
        return -1;
    }

    return 0;
}

static int
vzctDomainUpdateDeviceFlags(virDomainPtr dom,
                            const char *xml,
                            unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm = NULL;
    int ret = -1;
    vzctDriverConfigPtr cfg = NULL;
    virDomainDeviceDefPtr dev_cfg = NULL, dev_live = NULL;
    virCapsPtr caps = NULL;
    virDomainDefPtr vmdef = NULL;
    unsigned int parse_flags = VIR_DOMAIN_DEF_PARSE_INACTIVE |
                               VIR_DOMAIN_DEF_PARSE_ABI_UPDATE;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainUpdateDeviceFlagsEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (virDomainObjUpdateModificationImpact(vm, &flags) < 0)
        goto cleanup;

    cfg = vzctDriverGetConfig(driver);
    caps = vzctDriverGetCapabilities(driver);

    if (flags & VIR_DOMAIN_AFFECT_LIVE) {
        if (!(dev_live = virDomainDeviceDefParse(xml, vm->def,
                                                 caps, driver->xmlopt,
                                                 parse_flags)))
            goto cleanup;

        if (vzctDomainObjUpdateDevice(driver, vm, vm->def, dev_live, true) < 0)
            goto cleanup;

        if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm,
                                driver->caps) < 0)
            goto cleanup;

        /* no rollback in case saving status was failed, by design */
    }

    if (flags & VIR_DOMAIN_AFFECT_CONFIG) {
        if (!(dev_cfg = virDomainDeviceDefParse(xml, vm->def,
                                                caps, driver->xmlopt,
                                                parse_flags)))
            goto cleanup;

        if (!(vmdef = virDomainObjCopyPersistentDef(vm, caps, driver->xmlopt)))
            goto cleanup;

        if (vzctDomainObjUpdateDevice(driver, vm, vmdef, dev_cfg, false) < 0)
            goto cleanup;

        if (virDomainSaveConfig(cfg->configDir, driver->caps, vmdef) < 0)
            goto cleanup;

        virDomainObjAssignDef(vm, vmdef, false, NULL);
        vmdef = NULL;
    }

    ret = 0;

 cleanup:
    virDomainDefFree(vmdef);
    virDomainDeviceDefFree(dev_live);
    virDomainDeviceDefFree(dev_cfg);
    virObjectUnref(cfg);
    virObjectUnref(caps);
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);

    return ret;
}

static int
vzctDomainShutdownFlags(virDomainPtr dom, unsigned int flags)
{
    virDomainObjPtr vm;
    int ret = -1;
    stop_mode_e stopMode = M_HALT;
    vzctDriverPtr driver = dom->conn->privateData;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_SHUTDOWN_INITCTL |
                  VIR_DOMAIN_SHUTDOWN_SIGNAL, -1);

    if ((flags & VIR_DOMAIN_SHUTDOWN_INITCTL) &&
        (flags & VIR_DOMAIN_SHUTDOWN_SIGNAL)) {
        virReportError(VIR_ERR_INVALID_ARG,
                       _("Conflicting flags for virDomainShutdownFlagValues: %X"),
                       flags);
        return -1;
    }

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainShutdownFlagsEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (flags & VIR_DOMAIN_SHUTDOWN_SIGNAL)
        stopMode = M_KILL;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Domain is not running"));
        goto cleanup;
    }

    if (vzctlDomainDestroy(vm, stopMode) < 0)
        goto cleanup;

    vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_SHUTDOWN);
    ret = 0;

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainShutdown(virDomainPtr dom)
{
    return vzctDomainShutdownFlags(dom, 0);
}

static int
vzctDomainReboot(virDomainPtr dom, unsigned int flags)
{
    virDomainObjPtr vm;
    int ret = -1;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_REBOOT_INITCTL, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainRebootEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Domain is not running"));
        goto cleanup;
    }

    if (vzctlDomainRestart(vm) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;

}

static int
vzctDomainReset(virDomainPtr dom, unsigned int flags)
{
    virDomainObjPtr vm;
    int ret = -1;
    bool job = false;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainResetEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Domain is not running"));
        goto cleanup;
    }

    /* FIXME handle cases when domain can not start but already stopped */
    if (vzctlDomainReset(vm) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}


static int
vzctConnectGetCPUModelNames(virConnectPtr conn,
                            const char *archName,
                            char ***models,
                            unsigned int flags)
{
    virArch arch;

    virCheckFlags(0, -1);
    if (virConnectGetCPUModelNamesEnsureACL(conn) < 0)
        return -1;

    if (!(arch = virArchFromString(archName))) {
        virReportError(VIR_ERR_INVALID_ARG,
                       _("cannot find architecture %s"),
                       archName);
        return -1;
    }

    return virCPUGetModels(arch, models);
}

static const char *
vzctConnectGetType(virConnectPtr conn)
{
    if (virConnectGetTypeEnsureACL(conn) < 0)
        return NULL;

    return "vzct";
}

static int
vzctDomainGetAutostart(virDomainPtr dom, int *autostart)
{
    virDomainObjPtr vm;
    int ret = -1;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainGetAutostartEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    *autostart = vm->autostart;
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainSetAutostart(virDomainPtr dom, int autostart)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm;
    char *configFile = NULL, *autostartLink = NULL;
    int ret = -1;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    bool job = false;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainSetAutostartEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!vm->persistent) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("Cannot set autostart for transient domain"));
        goto cleanup;
    }

    autostart = (autostart != 0);

    if (vm->autostart == autostart) {
        ret = 0;
        goto cleanup;
    }

    configFile = virDomainConfigFile(cfg->configDir,
                                     vm->def->name);
    if (configFile == NULL)
        goto cleanup;
    autostartLink = virDomainConfigFile(cfg->autostartDir,
                                        vm->def->name);
    if (autostartLink == NULL)
        goto cleanup;

    if (autostart) {
        if (virFileMakePath(cfg->autostartDir) < 0) {
            virReportSystemError(errno,
                                 _("Cannot create autostart directory %s"),
                                 cfg->autostartDir);
            goto cleanup;
        }

        if (symlink(configFile, autostartLink) < 0) {
            virReportSystemError(errno,
                                 _("Failed to create symlink '%s to '%s'"),
                                 autostartLink, configFile);
            goto cleanup;
        }
    } else {
        if (unlink(autostartLink) < 0 && errno != ENOENT && errno != ENOTDIR) {
            virReportSystemError(errno,
                                 _("Failed to delete symlink '%s'"),
                                 autostartLink);
            goto cleanup;
        }
    }

    vm->autostart = autostart;
    ret = 0;

 cleanup:
    VIR_FREE(configFile);
    VIR_FREE(autostartLink);
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    virObjectUnref(cfg);
    return ret;
}

static int
vzctProcessAutostartDomain(virDomainObjPtr vm,
                           void *opaque ATTRIBUTE_UNUSED)
{
    int ret = -1;

    virObjectLock(vm);

    if (!vm->autostart ||
        virDomainObjIsActive(vm)) {

        ret = 0;
        goto cleanup;
    }

    if ((ret = vzctDomainObjBoot(vzct_driver, vm, false)) < 0) {
        VIR_ERROR(_("Failed to autostart VM '%s': %s"),
                  vm->def->name,
                  virGetLastErrorMessage());
    }

 cleanup:
    virObjectUnlock(vm);
    return ret;
}

static void
vzctStateAutoStart(void)
{
    if (!vzct_driver)
        return;

    virDomainObjListForEach(vzct_driver->domains,
                            vzctProcessAutostartDomain,
                            NULL);
}

static char *
vzctDomainGetOSType(virDomainPtr domain)
{
    virDomainObjPtr dom;
    char *ret = NULL;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainGetOSTypeEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ignore_value(VIR_STRDUP(ret, virDomainOSTypeToString(dom->def->os.type)));

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}

static int
vzctDomainGetVcpusFlags(virDomainPtr domain,
                        unsigned int flags)
{
    virDomainObjPtr dom;
    virDomainDefPtr def;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_DOMAIN_VCPU_MAXIMUM, -1);

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetVcpusFlagsEnsureACL(domain->conn, dom->def, flags) < 0)
        goto cleanup;

    if (!(def = virDomainObjGetOneDef(dom, flags)))
        goto cleanup;

    if (flags & VIR_DOMAIN_VCPU_MAXIMUM)
        ret = virDomainDefGetVcpusMax(dom->def);
    else
        ret = virDomainDefGetVcpus(dom->def);

 cleanup:
    virDomainObjEndAPI(&dom);

    return ret;
}

static int
vzctDomainGetMaxVcpus(virDomainPtr domain)
{
    return vzctDomainGetVcpusFlags(domain, (VIR_DOMAIN_AFFECT_LIVE |
                                            VIR_DOMAIN_VCPU_MAXIMUM));
}

static unsigned long long
vzctDomainGetMaxMemory(virDomainPtr domain)
{
    virDomainObjPtr dom = NULL;
    int ret = -1;

    if (!(dom = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainGetMaxMemoryEnsureACL(domain->conn, dom->def) < 0)
        goto cleanup;

    ret = virDomainDefGetMemoryTotal(dom->def);

 cleanup:
    virDomainObjEndAPI(&dom);
    return ret;
}

static int
vzctConnectSupportsFeature(virConnectPtr conn ATTRIBUTE_UNUSED, int feature)
{
    if (virConnectSupportsFeatureEnsureACL(conn) < 0)
        return -1;

    switch (feature) {
    case VIR_DRV_FEATURE_MIGRATION_PARAMS:
    case VIR_DRV_FEATURE_MIGRATION_P2P:
    case VIR_DRV_FEATURE_MIGRATE_CHANGE_PROTECTION:
        return 1;
    default:
        return 0;
    }
}

static int
vzctDomainSetVcpusFlags(virDomainPtr dom,
                        unsigned int nvcpus,
                        unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    virDomainObjPtr vm = NULL;
    virDomainDefPtr def;
    virDomainDefPtr persistentDef;
    int ret = -1;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_DOMAIN_VCPU_MAXIMUM, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainSetVcpusFlagsEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (virDomainObjGetDefs(vm, flags, &def, &persistentDef) < 0)
        goto cleanup;

    if (flags & VIR_DOMAIN_AFFECT_LIVE) {
        /* this configuration can be supported its just not used by clients
         * that deal with maximum cpu count value */
        if (flags & VIR_DOMAIN_VCPU_MAXIMUM) {
            virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                           _("maximum vcpu count of a live domain can't be modified"));
            goto cleanup;
        }

        /* support clients that do not care for maximum value */
        if (virDomainDefGetVcpusMax(def) < nvcpus &&
            virDomainDefSetVcpusMax(def, nvcpus, driver->xmlopt) < 0)
            goto cleanup;

        virDomainDefSetVcpus(def, nvcpus);

        if (vzctlSetVcpus(vm, nvcpus) < 0)
            goto cleanup;

        if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm, driver->caps) < 0)
            goto cleanup;
    }

    if (flags & VIR_DOMAIN_AFFECT_CONFIG) {
        if (flags & VIR_DOMAIN_VCPU_MAXIMUM) {

            if (virDomainDefSetVcpusMax(persistentDef, nvcpus, driver->xmlopt) < 0)
                goto cleanup;
        } else {
            /* support clients that do not care for maximum value */
            if (virDomainDefGetVcpusMax(persistentDef) < nvcpus &&
                virDomainDefSetVcpusMax(persistentDef, nvcpus, driver->xmlopt) < 0)
                goto cleanup;

            virDomainDefSetVcpus(persistentDef, nvcpus);
        }

        if (virDomainSaveConfig(cfg->configDir, driver->caps, persistentDef) < 0)
            goto cleanup;
    }

    ret = 0;
 cleanup:
    /* FIXME: rollback in case of error */
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainSetVcpus(virDomainPtr dom, unsigned int nvcpus)
{
    return vzctDomainSetVcpusFlags(dom, nvcpus, VIR_DOMAIN_AFFECT_LIVE);
}

static int
vzctDomainSetMemoryFlags(virDomainPtr dom,
                         unsigned long newmem,
                         unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
    virDomainObjPtr vm = NULL;
    virDomainDefPtr def;
    virDomainDefPtr persistentDef;
    int ret = -1;
    bool job = false;


    virCheckFlags(VIR_DOMAIN_AFFECT_LIVE |
                  VIR_DOMAIN_AFFECT_CONFIG |
                  VIR_DOMAIN_MEM_MAXIMUM, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainSetMemoryFlagsEnsureACL(dom->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (virDomainObjGetDefs(vm, flags, &def, &persistentDef) < 0)
        goto cleanup;

    if (flags & VIR_DOMAIN_AFFECT_LIVE) {
        /* this configuration can be supported its just not used by clients
         * that deal with maximum memory value */
        if (flags & VIR_DOMAIN_MEM_MAXIMUM) {
            virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                           _("cannot resize the maximum memory on an "
                             "active domain"));
            goto cleanup;
        }

        if (vzctlSetMemory(vm, newmem) < 0)
            goto cleanup;

        if (virDomainDefGetMemoryTotal(def) < newmem)
            virDomainDefSetMemoryTotal(def, newmem);
        def->mem.cur_balloon = newmem;

        if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm, driver->caps) < 0)
            goto cleanup;
    }

    if (flags & VIR_DOMAIN_AFFECT_CONFIG) {
        if (flags & VIR_DOMAIN_MEM_MAXIMUM) {
            virDomainDefSetMemoryTotal(persistentDef, newmem);
        } else {
            if (virDomainDefGetMemoryTotal(persistentDef) < newmem)
                virDomainDefSetMemoryTotal(persistentDef, newmem);
            persistentDef->mem.cur_balloon = newmem;
        }

        if (virDomainSaveConfig(cfg->configDir, driver->caps, persistentDef) < 0)
            goto cleanup;
    }

    ret = 0;
 cleanup:
    /* FIXME: rollback in case of error */
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainSetMemory(virDomainPtr dom, unsigned long newmem)
{
    return vzctDomainSetMemoryFlags(dom, newmem, VIR_DOMAIN_AFFECT_LIVE);
}

static int
vzctDomainSetMaxMemory(virDomainPtr dom, unsigned long memory)
{
    return vzctDomainSetMemoryFlags(dom, memory, VIR_DOMAIN_MEM_MAXIMUM);
}

static int
vzctDomainManagedSaveRemove(virDomainPtr dom, unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm;
    int ret = -1;
    char *dumppath = NULL;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainManagedSaveRemoveEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (!(dumppath = vzctDomainManagedSavePath(driver, vm)))
        goto cleanup;

    if (virFileDeleteTree(dumppath) < 0) {
        VIR_WARN("Failed to remove the managed state %s", dumppath);
        goto cleanup;
    }

    vm->hasManagedSave = false;
    ret = 0;

 cleanup:
    VIR_FREE(dumppath);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainHasManagedSaveImage(virDomainPtr dom, unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainHasManagedSaveImageEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    ret = vm->hasManagedSave;

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainManagedSave(virDomainPtr dom, unsigned int flags)
{
    vzctDriverPtr driver = dom->conn->privateData;
    virDomainObjPtr vm = NULL;
    int ret = -1;
    char *dumppath = NULL;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_SAVE_RUNNING, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainManagedSaveEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if (!vm->persistent) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("cannot do managed save for transient domain"));
        goto cleanup;
    }

    if (!(dumppath = vzctDomainManagedSavePath(driver, vm)))
        goto cleanup;

    if (vzctlDomainSuspend(vm, dumppath) < 0)
        goto cleanup;

    vm->hasManagedSave = true;
    vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_SAVED);
    ret = 0;

 cleanup:
    VIR_FREE(dumppath);
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainSnapshotWriteMetadata(virDomainObjPtr vm,
                                virDomainSnapshotObjPtr snapshot,
                                virCapsPtr caps,
                                virDomainXMLOptionPtr xmlopt,
                                char *snapshotDir)
{
    char *newxml = NULL;
    int ret = -1;
    char *snapDir = NULL;
    char *snapFile = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virUUIDFormat(vm->def->uuid, uuidstr);
    newxml = virDomainSnapshotDefFormat(
        uuidstr, snapshot->def, caps, xmlopt,
        virDomainDefFormatConvertXMLFlags(VIR_DOMAIN_XML_SECURE|VIR_DOMAIN_XML_UPDATE_CPU),
        1);
    if (newxml == NULL)
        return -1;

    if (virAsprintf(&snapDir, "%s/%s", snapshotDir, vm->def->name) < 0)
        goto cleanup;
    if (virFileMakePath(snapDir) < 0) {
        virReportSystemError(errno, _("cannot create snapshot directory '%s'"),
                             snapDir);
        goto cleanup;
    }

    if (virAsprintf(&snapFile, "%s/%s.xml", snapDir, snapshot->def->name) < 0)
        goto cleanup;

    ret = virXMLSaveFile(snapFile, NULL, "snapshot-edit", newxml);

 cleanup:
    VIR_FREE(snapFile);
    VIR_FREE(snapDir);
    VIR_FREE(newxml);
    return ret;
}

static virDomainObjPtr
vzctDomObjFromSnapshot(virDomainSnapshotPtr snapshot)
{
    return vzctDomObjFromDomain(snapshot->domain);
}

/* Looks up snapshot object from VM and name */
static virDomainSnapshotObjPtr
vzctSnapObjFromName(virDomainObjPtr vm,
                    const char *name)
{
    virDomainSnapshotObjPtr snap = NULL;
    snap = virDomainSnapshotFindByName(vm->snapshots, name);
    if (!snap)
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT,
                       _("no domain snapshot with matching name '%s'"),
                       name);

    return snap;
}

/* Looks up snapshot object from VM and snapshotPtr */
static virDomainSnapshotObjPtr
vzctSnapObjFromSnapshot(virDomainObjPtr vm,
                        virDomainSnapshotPtr snapshot)
{
    return vzctSnapObjFromName(vm, snapshot->name);
}


static virDomainSnapshotPtr
vzctDomainSnapshotCreateXML(virDomainPtr domain,
                            const char *xmlDesc,
                            unsigned int flags)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr vm = NULL;
    char *xml = NULL;
    virDomainSnapshotObjPtr snap = NULL;
    virDomainSnapshotPtr snapshot = NULL;
    virDomainSnapshotDefPtr def = NULL;
    bool update_current = true;
    unsigned int parse_flags = VIR_DOMAIN_SNAPSHOT_PARSE_DISKS;
    virDomainSnapshotObjPtr other = NULL;
    vzctDriverConfigPtr cfg = NULL;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_CREATE_ATOMIC |
                  VIR_DOMAIN_SNAPSHOT_CREATE_LIVE, NULL);

    if (!(vm = vzctDomObjFromDomain(domain)))
        goto cleanup;

    cfg = vzctDriverGetConfig(driver);

    if (virDomainSnapshotCreateXMLEnsureACL(domain->conn, vm->def, flags) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!(def = virDomainSnapshotDefParseString(xmlDesc, driver->caps, driver->xmlopt,
                                                parse_flags)))
        goto cleanup;


    if (!(xml = virDomainDefFormat(vm->def, driver->caps, 0)))
        goto endjob;

    if (!(def->dom = virDomainDefParseString(xml, driver->caps, driver->xmlopt, NULL,
                                         VIR_DOMAIN_DEF_PARSE_INACTIVE |
                                         VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE))) {
        goto endjob;
    }


    def->state = virDomainObjGetState(vm, NULL);

    if (virDomainObjIsActive(vm) &&
        def->memory == VIR_DOMAIN_SNAPSHOT_LOCATION_NONE) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("internal snapshot of a running VM "
                         "must include the memory state"));
        goto endjob;
    }

    def->memory = (def->state == VIR_DOMAIN_SHUTOFF ?
                   VIR_DOMAIN_SNAPSHOT_LOCATION_NONE :
                   VIR_DOMAIN_SNAPSHOT_LOCATION_INTERNAL);

    if (!(snap = virDomainSnapshotAssignDef(vm->snapshots, def)))
        goto endjob;

    def = NULL;

    if (update_current)
        snap->def->current = true;
    if (vm->current_snapshot) {
        if (VIR_STRDUP(snap->def->parent, vm->current_snapshot->def->name) < 0)
                goto endjob;
        if (update_current) {
            vm->current_snapshot->def->current = false;
            if (vzctDomainSnapshotWriteMetadata(vm, vm->current_snapshot,
                                                driver->caps, driver->xmlopt,
                                                cfg->snapshotDir) < 0)
                goto endjob;
            vm->current_snapshot = NULL;
        }
    }

    if (vzctlCreateSnapshot(vm, snap) < 0)
        goto endjob;

    /* If we fail after this point, there's not a whole lot we can
     * do; we've successfully taken the snapshot, and we are now running
     * on it, so we have to go forward the best we can
     */
    snapshot = virGetDomainSnapshot(domain, snap->def->name);

 endjob:
    if (snapshot) {
        if (vzctDomainSnapshotWriteMetadata(vm, snap, driver->caps,
                                            driver->xmlopt,
                                            cfg->snapshotDir) < 0) {
            /* if writing of metadata fails, error out rather than trying
             * to silently carry on without completing the snapshot */
            virObjectUnref(snapshot);
            snapshot = NULL;
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("unable to save metadata for snapshot %s"),
                           snap->def->name);
            virDomainSnapshotObjListRemove(vm->snapshots, snap);
        } else {
            if (update_current)
                vm->current_snapshot = snap;
            other = virDomainSnapshotFindByName(vm->snapshots,
                                                snap->def->parent);
            snap->parent = other;
            other->nchildren++;
            snap->sibling = other->first_child;
            other->first_child = snap;
        }
    } else if (snap) {
        virDomainSnapshotObjListRemove(vm->snapshots, snap);
    }

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    virDomainSnapshotDefFree(def);
    VIR_FREE(xml);
    virObjectUnref(cfg);
    return snapshot;
}

static char *
vzctDomainSnapshotGetXMLDesc(virDomainSnapshotPtr snapshot,
                             unsigned int flags)
{
    vzctDriverPtr driver = snapshot->domain->conn->privateData;
    virDomainObjPtr vm = NULL;
    char *xml = NULL;
    virDomainSnapshotObjPtr snap = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];

    virCheckFlags(VIR_DOMAIN_XML_SECURE, NULL);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return NULL;

    if (virDomainSnapshotGetXMLDescEnsureACL(snapshot->domain->conn, vm->def, flags) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    virUUIDFormat(snapshot->domain->uuid, uuidstr);

    xml = virDomainSnapshotDefFormat(uuidstr, snap->def, driver->caps,
                                     driver->xmlopt,
                                     virDomainDefFormatConvertXMLFlags(flags),
                                     0);
 cleanup:
    virDomainObjEndAPI(&vm);
    return xml;
}

static int
vzctDomainSnapshotNum(virDomainPtr domain,
                      unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(vm = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSnapshotNumEnsureACL(domain->conn, vm->def) < 0)
        goto cleanup;

    n = virDomainSnapshotObjListNum(vm->snapshots, NULL, flags);

 cleanup:
    virDomainObjEndAPI(&vm);
    return n;
}

static int
vzctDomainSnapshotListNames(virDomainPtr domain,
                            char **names,
                            int nameslen,
                            unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(vm = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainSnapshotListNamesEnsureACL(domain->conn, vm->def) < 0)
        goto cleanup;

    n = virDomainSnapshotObjListGetNames(vm->snapshots, NULL, names, nameslen,
                                         flags);

 cleanup:
    virDomainObjEndAPI(&vm);
    return n;
}

static int
vzctDomainListAllSnapshots(virDomainPtr domain,
                           virDomainSnapshotPtr **snaps,
                           unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_ROOTS |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(vm = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainListAllSnapshotsEnsureACL(domain->conn, vm->def) < 0)
        goto cleanup;

    n = virDomainListSnapshots(vm->snapshots, NULL, domain, snaps, flags);

 cleanup:
    virDomainObjEndAPI(&vm);
    return n;
}

static int
vzctDomainSnapshotNumChildren(virDomainSnapshotPtr snapshot,
                              unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    virDomainSnapshotObjPtr snap = NULL;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotNumChildrenEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    n = virDomainSnapshotObjListNum(vm->snapshots, snap, flags);

 cleanup:
    virDomainObjEndAPI(&vm);
    return n;
}

static int
vzctDomainSnapshotListChildrenNames(virDomainSnapshotPtr snapshot,
                                    char **names,
                                    int nameslen,
                                    unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    virDomainSnapshotObjPtr snap = NULL;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotListChildrenNamesEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    n = virDomainSnapshotObjListGetNames(vm->snapshots, snap, names, nameslen,
                                         flags);

 cleanup:
    virDomainObjEndAPI(&vm);
    return n;
}

static int
vzctDomainSnapshotListAllChildren(virDomainSnapshotPtr snapshot,
                                  virDomainSnapshotPtr **snaps,
                                  unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    virDomainSnapshotObjPtr snap = NULL;
    int n = -1;

    virCheckFlags(VIR_DOMAIN_SNAPSHOT_LIST_DESCENDANTS |
                  VIR_DOMAIN_SNAPSHOT_FILTERS_ALL, -1);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotListAllChildrenEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    n = virDomainListSnapshots(vm->snapshots, snap, snapshot->domain, snaps,
                               flags);

 cleanup:
    virDomainObjEndAPI(&vm);
    return n;
}

static virDomainSnapshotPtr
vzctDomainSnapshotLookupByName(virDomainPtr domain,
                               const char *name,
                               unsigned int flags)
{
    virDomainObjPtr vm;
    virDomainSnapshotObjPtr snap = NULL;
    virDomainSnapshotPtr snapshot = NULL;

    virCheckFlags(0, NULL);

    if (!(vm = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotLookupByNameEnsureACL(domain->conn, vm->def) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromName(vm, name)))
        goto cleanup;

    snapshot = virGetDomainSnapshot(domain, snap->def->name);

 cleanup:
    virDomainObjEndAPI(&vm);
    return snapshot;
}

static int
vzctDomainHasCurrentSnapshot(virDomainPtr domain,
                             unsigned int flags)
{
    virDomainObjPtr vm;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainHasCurrentSnapshotEnsureACL(domain->conn, vm->def) < 0)
        goto cleanup;

    ret = (vm->current_snapshot != NULL);

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static virDomainSnapshotPtr
vzctDomainSnapshotGetParent(virDomainSnapshotPtr snapshot,
                            unsigned int flags)
{
    virDomainObjPtr vm;
    virDomainSnapshotObjPtr snap = NULL;
    virDomainSnapshotPtr parent = NULL;

    virCheckFlags(0, NULL);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return NULL;

    if (virDomainSnapshotGetParentEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    if (!snap->def->parent) {
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT,
                       _("snapshot '%s' does not have a parent"),
                       snap->def->name);
        goto cleanup;
    }

    parent = virGetDomainSnapshot(snapshot->domain, snap->def->parent);

 cleanup:
    virDomainObjEndAPI(&vm);
    return parent;
}

static virDomainSnapshotPtr
vzctDomainSnapshotCurrent(virDomainPtr domain,
                          unsigned int flags)
{
    virDomainObjPtr vm;
    virDomainSnapshotPtr snapshot = NULL;

    virCheckFlags(0, NULL);

    if (!(vm = vzctDomObjFromDomain(domain)))
        return NULL;

    if (virDomainSnapshotCurrentEnsureACL(domain->conn, vm->def) < 0)
        goto cleanup;

    if (!vm->current_snapshot) {
        virReportError(VIR_ERR_NO_DOMAIN_SNAPSHOT, "%s",
                       _("the domain does not have a current snapshot"));
        goto cleanup;
    }

    snapshot = virGetDomainSnapshot(domain, vm->current_snapshot->def->name);

 cleanup:
    virDomainObjEndAPI(&vm);
    return snapshot;
}

static int
vzctDomainSnapshotIsCurrent(virDomainSnapshotPtr snapshot,
                            unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    int ret = -1;
    virDomainSnapshotObjPtr snap = NULL;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotIsCurrentEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    ret = (vm->current_snapshot &&
           STREQ(snapshot->name, vm->current_snapshot->def->name));

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainSnapshotHasMetadata(virDomainSnapshotPtr snapshot,
                              unsigned int flags)
{
    virDomainObjPtr vm = NULL;
    int ret = -1;
    virDomainSnapshotObjPtr snap = NULL;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    if (virDomainSnapshotHasMetadataEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    /* XXX Someday, we should recognize internal snapshots in qcow2
     * images that are not tied to a libvirt snapshot; if we ever do
     * that, then we would have a reason to return 0 here.  */
    ret = 1;

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainRevertToSnapshot(virDomainSnapshotPtr snapshot,
                           unsigned int flags)
{
    vzctDriverPtr driver = snapshot->domain->conn->privateData;
    virDomainObjPtr vm = NULL;
    int ret = -1;
    virDomainSnapshotObjPtr snap = NULL;
    virObjectEventPtr event = NULL;
    virObjectEventPtr event2 = NULL;
    int detail;
    virDomainDefPtr config = NULL;
    vzctDriverConfigPtr cfg = NULL;
    bool job = false;

    virCheckFlags(0, -1);

    /* We have the following transitions, which create the following events:
     * 1. inactive -> inactive: none
     * 2. inactive -> running:  EVENT_STARTED
     * 4. running  -> inactive: EVENT_STOPPED
     * 5. running  -> running:  none
     * 7. paused   -> inactive: EVENT_STOPPED
     * 8. paused   -> running:  EVENT_RESUMED
     * Also, several transitions occur even if we fail partway through,
     * and use of FORCE can cause multiple transitions.
     */

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    cfg = vzctDriverGetConfig(driver);

    if (virDomainRevertToSnapshotEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    if (!vm->persistent && snap->def->state != VIR_DOMAIN_RUNNING) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("transient domain needs to request run "
                         "to revert to inactive snapshot"));
        goto cleanup;
    }

    if (vm->current_snapshot) {
        vm->current_snapshot->def->current = false;
        if (vzctDomainSnapshotWriteMetadata(vm, vm->current_snapshot,
                                            driver->caps, driver->xmlopt,
                                            cfg->snapshotDir) < 0)
            goto cleanup;
        vm->current_snapshot = NULL;
        /* XXX Should we restore vm->current_snapshot after this point
         * in the failure cases where we know there was no change?  */
    }

    /* Prepare to copy the snapshot inactive xml as the config of this
     * domain.
     *
     * XXX Should domain snapshots track live xml rather
     * than inactive xml?  */
    snap->def->current = true;
    if (snap->def->dom) {
        config = virDomainDefCopy(snap->def->dom, driver->caps,
                                  driver->xmlopt, NULL, true);
        if (!config)
            goto cleanup;
    }

    switch ((virDomainState) snap->def->state) {
    case VIR_DOMAIN_RUNNING:
        /* Transitions 2, 3, 5, 6, 8, 9 */
        /* When using the loadvm monitor command, vzct does not know
         * whether to pause or run the reverted domain, and just stays
         * in the same state as before the monitor command, whether
         * that is paused or running.  We always pause before loadvm,
         * to have finer control.  */
        if (virDomainObjIsActive(vm)) {
            /* Transitions 5, 6, 8, 9 */
            /* Check for ABI compatibility. We need to do this check against
             * the migratable XML or it will always fail otherwise */
            if (virDomainObjGetState(vm, NULL) == VIR_DOMAIN_PAUSED) {
                /* Transitions 5, 6 */
                detail = VIR_DOMAIN_EVENT_STARTED_FROM_SNAPSHOT;
                event = virDomainEventLifecycleNewFromObj(vm,
                                                 VIR_DOMAIN_EVENT_STARTED,
                                                 detail);
            }

            if (vzctlRevertSnapshot(vm, snap) < 0)
                goto cleanup;

            if (config)
                virDomainObjAssignDef(vm, config, false, NULL);
        } else {
            /* Transitions 2, 3 */
            if (config)
                virDomainObjAssignDef(vm, config, false, NULL);

            if (vzctlRevertSnapshot(vm, snap) < 0)
                goto cleanup;

            virDomainAuditStart(vm, "from-snapshot", 0);
            detail = VIR_DOMAIN_EVENT_STARTED_FROM_SNAPSHOT;
            event = virDomainEventLifecycleNewFromObj(vm,
                                             VIR_DOMAIN_EVENT_STARTED,
                                             detail);
        }

        virDomainObjSetState(vm, VIR_DOMAIN_RUNNING,
                             VIR_DOMAIN_RUNNING_FROM_SNAPSHOT);
        break;

    case VIR_DOMAIN_SHUTOFF:
        /* Transitions 1, 4, 7 */
        /* Newer vzct -loadvm refuses to revert to the state of a snapshot
         * created by vzct-img snapshot -c.  If the domain is running, we
         * must take it offline; then do the revert using vzct-img.
         */

        if (virDomainObjIsActive(vm)) {
            virDomainAuditStop(vm, "from-snapshot");
            detail = VIR_DOMAIN_EVENT_STOPPED_FROM_SNAPSHOT;
            event = virDomainEventLifecycleNewFromObj(vm,
                                             VIR_DOMAIN_EVENT_STOPPED,
                                             detail);
        }

        if (vzctlRevertSnapshot(vm, snap) < 0)
            goto cleanup;

        if (config)
            virDomainObjAssignDef(vm, config, false, NULL);

    default:
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Invalid target domain state '%s'. Refusing "
                         "snapshot reversion"),
                       virDomainStateTypeToString(snap->def->state));
        goto cleanup;
    }

    ret = 0;

 cleanup:
    if (ret == 0) {
        if (vzctDomainSnapshotWriteMetadata(vm, snap, driver->caps,
                                            driver->xmlopt, cfg->snapshotDir) < 0)
            ret = -1;
        else
            vm->current_snapshot = snap;
    } else if (snap) {
        snap->def->current = false;
    }
    if (ret == 0 && config && vm->persistent &&
        !(ret = virDomainSaveConfig(cfg->configDir, driver->caps,
                                    vm->newDef ? vm->newDef : vm->def))) {
        detail = VIR_DOMAIN_EVENT_DEFINED_FROM_SNAPSHOT;
        virObjectEventStateQueue(driver->domainEventState,
            virDomainEventLifecycleNewFromObj(vm,
                                              VIR_DOMAIN_EVENT_DEFINED,
                                              detail));
    }
    if (event) {
        virObjectEventStateQueue(driver->domainEventState, event);
        virObjectEventStateQueue(driver->domainEventState, event2);
    }
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    virObjectUnref(cfg);

    return ret;
}


typedef struct _vzctSnapReparent vzctSnapReparent;
typedef vzctSnapReparent *vzctSnapReparentPtr;
struct _vzctSnapReparent {
    vzctDriverConfigPtr cfg;
    virDomainSnapshotObjPtr parent;
    virDomainObjPtr vm;
    virCapsPtr caps;
    virDomainXMLOptionPtr xmlopt;
    int err;
    virDomainSnapshotObjPtr last;
};


static int
vzctDomainSnapshotReparentChildren(void *payload,
                                   const void *name ATTRIBUTE_UNUSED,
                                   void *data)
{
    virDomainSnapshotObjPtr snap = payload;
    vzctSnapReparentPtr rep = data;

    if (rep->err < 0)
        return 0;

    VIR_FREE(snap->def->parent);
    snap->parent = rep->parent;

    if (rep->parent->def &&
        VIR_STRDUP(snap->def->parent, rep->parent->def->name) < 0) {
        rep->err = -1;
        return 0;
    }

    if (!snap->sibling)
        rep->last = snap;

    rep->err = vzctDomainSnapshotWriteMetadata(rep->vm, snap, rep->caps,
                                               rep->xmlopt,
                                               rep->cfg->snapshotDir);
    return 0;
}

/* Discard one snapshot (or its metadata), without reparenting any children.  */
static int
vzctDomainSnapshotDiscard(vzctDriverPtr driver,
                          virDomainObjPtr vm,
                          virDomainSnapshotObjPtr snap)
{
    char *snapFile = NULL;
    int ret = -1;
    virDomainSnapshotObjPtr parentsnap = NULL;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);

    if (vzctlDeleteSnapshot(vm, snap) < 0)
        goto cleanup;

    if (virAsprintf(&snapFile, "%s/%s/%s.xml", cfg->snapshotDir,
                    vm->def->name, snap->def->name) < 0)
        goto cleanup;

    if (snap == vm->current_snapshot) {
        if (snap->def->parent) {
            parentsnap = virDomainSnapshotFindByName(vm->snapshots,
                                                     snap->def->parent);
            if (!parentsnap) {
                VIR_WARN("missing parent snapshot matching name '%s'",
                         snap->def->parent);
            } else {
                parentsnap->def->current = true;
                if (vzctDomainSnapshotWriteMetadata(vm, parentsnap, driver->caps,
                                                    driver->xmlopt,
                                                    cfg->snapshotDir) < 0) {
                    VIR_WARN("failed to set parent snapshot '%s' as current",
                             snap->def->parent);
                    parentsnap->def->current = false;
                    parentsnap = NULL;
                }
            }
        }
        vm->current_snapshot = parentsnap;
    }

    if (unlink(snapFile) < 0)
        VIR_WARN("Failed to unlink %s", snapFile);
    virDomainSnapshotObjListRemove(vm->snapshots, snap);

    ret = 0;

 cleanup:
    VIR_FREE(snapFile);
    virObjectUnref(cfg);
    return ret;
}

static int
vzctDomainSnapshotDelete(virDomainSnapshotPtr snapshot,
                         unsigned int flags)
{
    vzctDriverPtr driver = snapshot->domain->conn->privateData;
    virDomainObjPtr vm = NULL;
    int ret = -1;
    virDomainSnapshotObjPtr snap = NULL;
    vzctSnapReparent rep;
    vzctDriverConfigPtr cfg = NULL;
    bool job = false;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromSnapshot(snapshot)))
        return -1;

    cfg = vzctDriverGetConfig(driver);

    if (virDomainSnapshotDeleteEnsureACL(snapshot->domain->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!(snap = vzctSnapObjFromSnapshot(vm, snapshot)))
        goto cleanup;

    if (snap->nchildren) {
        rep.cfg = cfg;
        rep.parent = snap->parent;
        rep.vm = vm;
        rep.err = 0;
        rep.last = NULL;
        rep.caps = driver->caps;
        rep.xmlopt = driver->xmlopt;
        virDomainSnapshotForEachChild(snap,
                                      vzctDomainSnapshotReparentChildren,
                                      &rep);
        if (rep.err < 0)
            goto cleanup;
        /* Can't modify siblings during ForEachChild, so do it now.  */
        snap->parent->nchildren += snap->nchildren;
        rep.last->sibling = snap->parent->first_child;
        snap->parent->first_child = snap->first_child;
    }

    virDomainSnapshotDropParent(snap);
    ret = vzctDomainSnapshotDiscard(driver, vm, snap);

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    virObjectUnref(cfg);
    return ret;
}

#define VZCT_MIGRATION_FLAGS   VIR_MIGRATE_CHANGE_PROTECTION | \
                               VIR_MIGRATE_NON_SHARED_INC | \
                               VIR_MIGRATE_UNDEFINE_SOURCE | \
                               VIR_MIGRATE_PERSIST_DEST | \
                               VIR_MIGRATE_LIVE

#define VZCT_MIGRATION_PARAMETERS                               \
    VIR_MIGRATE_PARAM_DEST_XML,         VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_URI,              VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_DEST_NAME,        VIR_TYPED_PARAM_STRING, \
    VIR_MIGRATE_PARAM_MIGRATE_DISKS,    VIR_TYPED_PARAM_STRING |  \
                                        VIR_TYPED_PARAM_MULTIPLE, \
    VIR_MIGRATE_PARAM_BANDWIDTH,        VIR_TYPED_PARAM_ULLONG, \
    NULL

static char *
vzctDomainMigrateBeginStep(vzctDriverPtr driver, virDomainObjPtr vm)
{
    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("Only active domain migration is supported"));
        return NULL;
    }

    return virDomainDefFormat(vm->def, driver->caps,
                              VIR_DOMAIN_XML_MIGRATABLE |
                              VIR_DOMAIN_XML_SECURE |
                              VIR_DOMAIN_XML_INACTIVE);
}

static char *
vzctDomainMigrateBegin3Params(virDomainPtr dom,
                              virTypedParameterPtr params ATTRIBUTE_UNUSED,
                              int nparams ATTRIBUTE_UNUSED,
                              char **cookieout ATTRIBUTE_UNUSED,
                              int *cookieoutlen ATTRIBUTE_UNUSED,
                              unsigned int flags)
{
    char* xml = NULL;
    virDomainObjPtr vm;
    vzctDriverPtr driver = dom->conn->privateData;
    vzctDomainObjPrivatePtr priv;
    bool job = false;

    virCheckFlags(VZCT_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return NULL;

    if (virTypedParamsGet(params, nparams, VIR_MIGRATE_PARAM_DEST_XML)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("Changing destination XML is not supported"));
        return NULL;
    }

    if (!(vm = vzctDomObjFromDomain(dom)))
        return NULL;

    if (virDomainMigrateBegin3ParamsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;
    priv = vm->privateData;
    priv->migration = true;

    xml = vzctDomainMigrateBeginStep(driver, vm);

 cleanup:
    if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION) || !xml)
        vzctDomainObjEndJob(vm, &job);

    virDomainObjEndAPI(&vm);
    return xml;
}


static char*
vzctMigrationCreateURI(void)
{
    char *hostname = NULL;
    char *uri = NULL;

    if (!(hostname = virGetHostname()))
        goto cleanup;

    if (STRPREFIX(hostname, "localhost")) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("hostname on destination resolved to localhost,"
                         " but migration requires an FQDN"));
        goto cleanup;
    }

    if (virAsprintf(&uri, "vzctmigr://%s", hostname) < 0)
        goto cleanup;

 cleanup:
    VIR_FREE(hostname);
    return uri;
}

static int
vzctBakeCookie(vzctDriverPtr driver,
               virDomainObjPtr vm,
               bool persistent,
               char **cookieout,
               int *cookieoutlen)
{
    virBuffer buf = VIR_BUFFER_INITIALIZER;
    vzctDomainObjPrivatePtr priv = vm->privateData;
    virCapsPtr caps = vzctDriverGetCapabilities(driver);
    int ret = -1;

    *cookieout = NULL;
    *cookieoutlen = 0;

    virBufferAddLit(&buf, "<vzct-migration>\n");
    virBufferAdjustIndent(&buf, 2);
    if (priv->migrationPort)
        virBufferAsprintf(&buf, "<port>%d</port>\n", priv->migrationPort);
    if (persistent &&
        vm->persistent &&
        virDomainDefFormatInternal(vm->newDef, caps,
                                   VIR_DOMAIN_DEF_FORMAT_SECURE, &buf,
                                   driver->xmlopt) < 0)
        goto error;

    virBufferAdjustIndent(&buf, -2);
    virBufferAddLit(&buf, "</vzct-migration>");

    if (virBufferCheckError(&buf) < 0)
        goto cleanup;

    *cookieout = virBufferContentAndReset(&buf);
    *cookieoutlen = strlen(*cookieout) + 1;
    ret = 0;

 cleanup:
    virObjectUnref(caps);
    return ret;

 error:
    virBufferFreeAndReset(&buf);
    goto cleanup;
}

static int
vzctDomainMigratePrepare3Params(virConnectPtr conn,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein ATTRIBUTE_UNUSED,
                                int cookieinlen ATTRIBUTE_UNUSED,
                                char **cookieout,
                                int *cookieoutlen,
                                char **uri_out ATTRIBUTE_UNUSED,
                                unsigned int flags)
{
    int ret = -1;
    vzctDriverPtr driver = conn->privateData;
    virDomainDefPtr def = NULL;
    const char *xml;
    virCapsPtr caps = NULL;
    virDomainObjPtr vm = NULL;
    const char *uri = NULL;
    bool job = false;
    int nmigrate_disks;
    const char **migrate_disks = NULL;
    virObjectEventPtr event;
    vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);

    virCheckFlags(VZCT_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return -1;

    if (virTypedParamsGet(params, nparams, VIR_MIGRATE_PARAM_DEST_NAME)) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("Changing destination name is not supported"));
        return -1;
    }

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_XML, &xml) < 0 ||
        virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_URI, &uri) < 0)
        goto cleanup;

    nmigrate_disks = virTypedParamsGetStringList(params, nparams,
                                                 VIR_MIGRATE_PARAM_MIGRATE_DISKS,
                                                 &migrate_disks);
    if (nmigrate_disks < 0)
        goto cleanup;

    caps = vzctDriverGetCapabilities(driver);

    if (!(def = virDomainDefParseString(xml, caps, driver->xmlopt, NULL,
                                        VIR_DOMAIN_DEF_PARSE_INACTIVE)))
        goto cleanup;

    if (virDomainMigratePrepare3ParamsEnsureACL(conn, def) < 0)
        goto cleanup;

    if (!uri && !(*uri_out = vzctMigrationCreateURI()))
        goto cleanup;

    if (!(vm = virDomainObjListAdd(driver->domains, def,
                                   driver->xmlopt,
                                   VIR_DOMAIN_OBJ_LIST_ADD_CHECK_LIVE |
                                   VIR_DOMAIN_OBJ_LIST_ADD_LIVE,
                                   NULL)))
        goto cleanup;

    def = NULL;
    virObjectRef(vm);

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (vzctDomainObjStart(driver, vm, NULL, true) < 0)
        goto cleanup;

    if (vzctlMigrationStartServer(driver, vm,
                                  nmigrate_disks, migrate_disks,
                                  flags & VIR_MIGRATE_NON_SHARED_INC) < 0)
        goto cleanup;

    virDomainObjSetState(vm, VIR_DOMAIN_PAUSED,
                             VIR_DOMAIN_PAUSED_MIGRATION);
    if (virDomainSaveStatus(driver->xmlopt, cfg->stateDir, vm, driver->caps) < 0)
        goto cleanup;

    event = virDomainEventLifecycleNewFromObj(vm,
                                              VIR_DOMAIN_EVENT_STARTED,
                                              VIR_DOMAIN_EVENT_STARTED_MIGRATED);
    if (event)
        virObjectEventStateQueue(driver->domainEventState, event);

    vzctBakeCookie(driver, vm, false, cookieout, cookieoutlen);
    ret = 0;

 cleanup:

    if (ret < 0 && vm) {
        vzctlMigrationStopServer(driver, vm);
        vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_FAILED);
        vzctDomainObjEndJob(vm, &job);
    }

    virDomainDefFree(def);
    virObjectUnref(caps);
    virDomainObjEndAPI(&vm);
    virObjectUnref(cfg);

    return ret;
}

typedef struct _vzctMigrationCookie vzctMigrationCookie;
typedef vzctMigrationCookie *vzctMigrationCookiePtr;
struct _vzctMigrationCookie {
    int port;
    virDomainDefPtr persistent;
};

static void
vzctMigrationCookieFree(vzctMigrationCookiePtr mig)
{
    if (!mig)
        return;

    virDomainDefFree(mig->persistent);
    VIR_FREE(mig);
}

static vzctMigrationCookiePtr
vzctEatCookie(vzctDriverPtr driver, const char *cookiein, int cookieinlen)
{
    vzctMigrationCookiePtr mig;
    vzctMigrationCookiePtr ret = NULL;
    xmlDocPtr doc = NULL;
    xmlXPathContextPtr ctx = NULL;
    virCapsPtr caps;
    xmlNodePtr domnode;

    if (VIR_ALLOC(mig) < 0)
        return NULL;

    caps = vzctDriverGetCapabilities(driver);

    if (!cookiein || cookieinlen <= 0 || cookiein[cookieinlen - 1] != '\0') {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Invalid migration cookie"));
        goto cleanup;
    }

    if (!(doc = virXMLParseStringCtxt(cookiein,
                                      _("(_migration_cookie)"), &ctx)))
        goto cleanup;

    if (virXPathBoolean("count(./port) > 0", ctx) &&
        virXPathInt("string(./port[1])", ctx, &mig->port) < 0)
        goto cleanup;

    if ((domnode = virXPathNode("./domain[1]", ctx)) &&
        !(mig->persistent =
            virDomainDefParseNode(doc, domnode,
                                  caps, driver->xmlopt, NULL,
                                  VIR_DOMAIN_DEF_PARSE_INACTIVE |
                                  VIR_DOMAIN_DEF_PARSE_SKIP_VALIDATE)))
        goto cleanup;

    ret = mig;
    mig = NULL;

 cleanup:
    virObjectUnref(caps);
    VIR_FREE(mig);
    xmlXPathFreeContext(ctx);
    xmlFreeDoc(doc);

    return ret;
}

static int
vzctDomainMigratePerformStep(vzctDriverPtr driver,
                             virDomainObjPtr vm,
                             virTypedParameterPtr params,
                             int nparams,
                             const char *cookiein,
                             int cookieinlen,
                             char **cookieout,
                             int *cookieoutlen,
                             unsigned int flags)
{
    int ret = -1;
    const char *uri;
    vzctMigrationCookiePtr mig = NULL;
    vzctDomainObjPrivatePtr priv = vm->privateData;
    int rc;
    int status;
    int nmigrate_disks;
    const char **migrate_disks = NULL;
    unsigned long long bandwidth = 0;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_URI, &uri) < 0)
        goto cleanup;

    nmigrate_disks = virTypedParamsGetStringList(params, nparams,
                                                 VIR_MIGRATE_PARAM_MIGRATE_DISKS,
                                                 &migrate_disks);
    if (nmigrate_disks < 0)
        goto cleanup;

    if (virTypedParamsGetULLong(params, nparams, VIR_MIGRATE_PARAM_BANDWIDTH,
                                &bandwidth) < 0)
        goto cleanup;

    if (bandwidth > 0) {
        virReportError(VIR_ERR_OPERATION_UNSUPPORTED, "%s",
                       _("Bandwidth rate limiting is not supported"));
        goto cleanup;
    }

    if (!(mig = vzctEatCookie(driver, cookiein, cookieinlen)))
        goto cleanup;

    if (!mig->port) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s", _("missing port"));
        goto cleanup;
    }

    if (vzctlMigrationStartClient(vm, uri, mig->port,
                                  nmigrate_disks, migrate_disks,
                                  flags & VIR_MIGRATE_NON_SHARED_INC,
                                  flags & VIR_MIGRATE_LIVE) < 0)
        goto cleanup;

    virObjectUnlock(vm);
    rc = virProcessWait(priv->migrationPid, &status, true);
    virObjectLock(vm);

    if (rc < 0)
        goto cleanup;

    if (!(WIFEXITED(status) && WEXITSTATUS(status) == 0)) {
        if (priv->jobAborted) {
            virReportError(VIR_ERR_OPERATION_ABORTED, "%s",
                           _("migration cancelled by client"));
        } else {
            char *st = virProcessTranslateStatus(status);
            virReportError(VIR_ERR_INTERNAL_ERROR,
                           _("Migration process unexpected %s"), NULLSTR(st));
        }
        goto cleanup;
    }

    vzctBakeCookie(driver, vm,
                   flags & VIR_MIGRATE_PERSIST_DEST,
                   cookieout, cookieoutlen);

    ret = 0;

 cleanup:
    priv->jobAborted = false;
    priv->migrationPid = 0;
    vzctMigrationCookieFree(mig);

    return ret;
}

static void
vzctDomainMigrateConfirmStep(vzctDriverPtr driver,
                             virDomainObjPtr vm,
                             unsigned int flags)
{
    vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_MIGRATED);

    if (vm->persistent && (flags & VIR_MIGRATE_UNDEFINE_SOURCE)) {
        vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);

        if (virDomainDeleteConfig(cfg->configDir, cfg->autostartDir, vm) < 0)
            VIR_WARN("Cannot undefine domain %s", vm->def->name);
        else
            virDomainObjListRemove(driver->domains, vm);

        virObjectUnref(cfg);
    }
}

static int
vzctDomainMigratePerformP2P(vzctDriverPtr driver,
                            virDomainObjPtr vm,
                            const char *dconnuri,
                            virTypedParameterPtr orig_params,
                            int nparams,
                            unsigned int flags)
{
    int ret = -1;
    vzctDomainObjPrivatePtr priv = vm->privateData;
    virErrorPtr orig_err = NULL;
    char *uri = NULL;
    char *cookiein = NULL;
    char *cookieout = NULL;
    int cookieinlen = 0;
    int cookieoutlen = 0;
    virConnectPtr dconn = NULL;
    char *xml = NULL;
    virTypedParameterPtr params = NULL;
    int maxparams = nparams;
    bool job = false;

    if (virTypedParamsCopy(&params, orig_params, nparams) < 0)
        return -1;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;
    priv->migration = true;

    if (!(dconn = virConnectOpen(dconnuri)))
        goto cleanup;

    if (!(xml = vzctDomainMigrateBeginStep(driver, vm)))
        goto cleanup;

    if (virTypedParamsAddString(&params, &nparams, &maxparams,
                                VIR_MIGRATE_PARAM_DEST_XML, xml) < 0)
        goto cleanup;

    cookiein = cookieout;
    cookieinlen = cookieoutlen;
    cookieout = NULL;
    cookieoutlen = 0;

    if (dconn->driver->domainMigratePrepare3Params
            (dconn, params, nparams, cookiein, cookieinlen,
             &cookieout, &cookieoutlen, &uri, flags) < 0)
        goto cleanup;

    if (uri && virTypedParamsReplaceString(&params, &nparams,
                                           VIR_MIGRATE_PARAM_URI, uri) < 0) {
        orig_err = virSaveLastError();
        goto finish;
    }

    VIR_FREE(cookiein);
    cookiein = cookieout;
    cookieinlen = cookieoutlen;
    cookieout = NULL;
    cookieoutlen = 0;

    if (vzctDomainMigratePerformStep(driver, vm, params, nparams,
                                     cookiein, cookieinlen,
                                     &cookieout, &cookieoutlen, flags) < 0) {
        orig_err = virSaveLastError();
        goto finish;
    }

    ret = 0;

 finish:
    VIR_FREE(cookiein);
    cookiein = cookieout;
    cookieinlen = cookieoutlen;
    cookieout = NULL;
    cookieoutlen = 0;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_NAME, NULL) == 1 ||
        virTypedParamsReplaceString(&params, &nparams,
                                    VIR_MIGRATE_PARAM_DEST_NAME,
                                    vm->def->name) == 0) {
        virDomainPtr ddomain;
        ddomain = dconn->driver->domainMigrateFinish3Params(dconn, params, nparams,
                                                            cookiein, cookieinlen,
                                                            NULL, NULL,
                                                            flags,
                                                            ret == 0 ? 0 : 1);
        virObjectUnref(ddomain);
    }

    if (ret == 0)
        vzctDomainMigrateConfirmStep(driver, vm, flags);

 cleanup:
    if (ret < 0 && !orig_err)
        orig_err = virSaveLastError();

    virTypedParamsFree(params, nparams);
    virObjectUnref(dconn);
    VIR_FREE(xml);
    VIR_FREE(uri);
    VIR_FREE(cookiein);
    VIR_FREE(cookieout);
    vzctDomainObjEndJob(vm, &job);

    if (orig_err) {
        virSetError(orig_err);
        virFreeError(orig_err);
    }

    return ret;
}

static int
vzctDomainMigratePerform3Params(virDomainPtr dom,
                                const char *dconnuri,
                                virTypedParameterPtr params,
                                int nparams,
                                const char *cookiein,
                                int cookieinlen,
                                char **cookieout,
                                int *cookieoutlen,
                                unsigned int flags)
{
    int ret = -1;
    virDomainObjPtr vm;
    vzctDriverPtr driver = dom->conn->privateData;

    virCheckFlags(VZCT_MIGRATION_FLAGS | VIR_MIGRATE_PEER2PEER, -1);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return -1;

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainMigratePerform3ParamsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (flags & VIR_MIGRATE_PEER2PEER) {
        ret = vzctDomainMigratePerformP2P(driver, vm, dconnuri,
                                          params, nparams,
                                          flags & (~VIR_MIGRATE_PEER2PEER));
    } else {
        if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION) &&
            vzctDomainObjBeginJob(vm, NULL) < 0)
            goto cleanup;

        ret = vzctDomainMigratePerformStep(driver, vm, params, nparams,
                                           cookiein, cookieinlen,
                                           cookieout, cookieoutlen,
                                           flags);

        if (!(flags & VIR_MIGRATE_CHANGE_PROTECTION) || ret < 0)
            vzctDomainObjEndJob(vm, NULL);
    }

 cleanup:
    virDomainObjEndAPI(&vm);

    return ret;
}

static virDomainPtr
vzctDomainMigrateFinish3Params(virConnectPtr dconn,
                               virTypedParameterPtr params,
                               int nparams,
                               const char *cookiein,
                               int cookieinlen,
                               char **cookieout ATTRIBUTE_UNUSED,
                               int *cookieoutlen ATTRIBUTE_UNUSED,
                               unsigned int flags,
                               int cancelled)
{
    vzctDriverPtr driver = dconn->privateData;
    const char *name = NULL;
    virDomainObjPtr vm;
    virDomainPtr dom = NULL;
    virObjectEventPtr event;

    virCheckFlags(VZCT_MIGRATION_FLAGS, NULL);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return NULL;

    if (virTypedParamsGetString(params, nparams,
                                VIR_MIGRATE_PARAM_DEST_NAME,
                                &name) < 0)
        return NULL;

    if (!name) {
        virReportError(VIR_ERR_NO_DOMAIN, "%s", _("missing domain name"));
        return NULL;
    }

    if (!(vm = virDomainObjListFindByName(driver->domains, name))) {
        virReportError(VIR_ERR_NO_DOMAIN,
                       _("no domain with matching name '%s'"), name);
        return NULL;
    }

    if (virDomainMigrateFinish3ParamsEnsureACL(dconn, vm->def) < 0) {
        virDomainObjEndAPI(&vm);
        return NULL;
    }

    if (cancelled) {
        vzctDomainStop(driver, vm, VIR_DOMAIN_SHUTOFF_FAILED);
        /* need to report some error */
        virReportError(VIR_ERR_MIGRATE_FINISH_OK, NULL);
        goto cleanup;
    }

    event = virDomainEventLifecycleNewFromObj(vm,
                                              VIR_DOMAIN_EVENT_RESUMED,
                                              VIR_DOMAIN_EVENT_RESUMED_MIGRATED);
    if (event)
        virObjectEventStateQueue(driver->domainEventState, event);

    virDomainObjSetState(vm, VIR_DOMAIN_RUNNING, VIR_DOMAIN_RUNNING_MIGRATED);

    if (flags & VIR_MIGRATE_PERSIST_DEST) {
        vzctMigrationCookiePtr mig = vzctEatCookie(driver,
                                                   cookiein, cookieinlen);
        vzctDriverConfigPtr cfg = vzctDriverGetConfig(driver);
        virDomainDefPtr newdef = NULL;
        virCapsPtr caps = vzctDriverGetCapabilities(driver);

        if (mig) {
            if (mig->persistent)
                VIR_STEAL_PTR(newdef, mig->persistent);
            else
                newdef = virDomainDefCopy(vm->def, caps, driver->xmlopt, NULL, false);
        }

        if (!newdef || virDomainSaveConfig(cfg->configDir, driver->caps, newdef) < 0) {
            VIR_WARN("Cannot persist domain %s", vm->def->name);
        } else {
            virDomainDefFree(vm->newDef);
            VIR_STEAL_PTR(vm->newDef, newdef);
            vm->persistent = true;
        }

        virObjectUnref(caps);
        virObjectUnref(cfg);
        vzctMigrationCookieFree(mig);
        virDomainDefFree(newdef);
    }

    dom = virGetDomain(dconn, vm->def->name, vm->def->uuid, vm->def->id);

 cleanup:
    vzctlMigrationStopServer(driver, vm);
    vzctDomainObjEndJob(vm, NULL);
    virDomainObjEndAPI(&vm);

    return dom;
}

static int
vzctDomainMigrateConfirm3Params(virDomainPtr domain,
                                virTypedParameterPtr params ATTRIBUTE_UNUSED,
                                int nparams ATTRIBUTE_UNUSED,
                                const char *cookiein ATTRIBUTE_UNUSED,
                                int cookieinlen ATTRIBUTE_UNUSED,
                                unsigned int flags,
                                int cancelled)
{
    vzctDriverPtr driver = domain->conn->privateData;
    virDomainObjPtr vm;

    virCheckFlags(VZCT_MIGRATION_FLAGS, -1);

    if (virTypedParamsValidate(params, nparams, VZCT_MIGRATION_PARAMETERS) < 0)
        return -1;

    if (!(vm = vzctDomObjFromDomain(domain)))
        return -1;

    if (virDomainMigrateConfirm3ParamsEnsureACL(domain->conn, vm->def) < 0) {
        virDomainObjEndAPI(&vm);
        return -1;
    }

    if (!cancelled)
        vzctDomainMigrateConfirmStep(driver, vm, flags);

    if (flags & VIR_MIGRATE_CHANGE_PROTECTION)
        vzctDomainObjEndJob(vm, NULL);

    virDomainObjEndAPI(&vm);
    return 0;
}

static int
vzctDomainAbortJob(virDomainPtr dom)
{
    virDomainObjPtr vm = NULL;
    vzctDomainObjPrivatePtr priv;
    int ret = -1;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainAbortJobEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("domain is not running"));
        goto cleanup;
    }


    priv = vm->privateData;
    if (!priv->migrationPid) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("no job is active on the domain"));
        goto cleanup;
    }

    virProcessKill(priv->migrationPid, SIGTERM);
    priv->jobAborted = true;
    ret = 0;

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainSetUserPassword(virDomainPtr dom,
                          const char *user,
                          const char *password,
                          unsigned int flags)
{
    virDomainObjPtr vm;
    int ret = -1;
    bool job = false;

    virCheckFlags(VIR_DOMAIN_PASSWORD_ENCRYPTED, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainSetUserPasswordEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    ret = vzctlDomainSetUserPassword(vm, user, password, flags);

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainBlockResize(virDomainPtr dom,
                      const char *path,
                      unsigned long long size,
                      unsigned int flags)
{
    virDomainObjPtr vm;
    virDomainDiskDefPtr disk;
    bool job;
    int ret = -1;

    virCheckFlags(VIR_DOMAIN_BLOCK_RESIZE_BYTES, -1);

    if (path[0] == '\0') {
        virReportError(VIR_ERR_INVALID_ARG,
                       "%s", _("empty path"));
        return -1;
    }

    if (flags & VIR_DOMAIN_BLOCK_RESIZE_BYTES)
        size >>= 10;

    if (!(vm = vzctDomObjFromDomain(dom)))
        goto cleanup;

    if (virDomainBlockResizeEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainObjBeginJob(vm, &job) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID,
                       "%s", _("domain is not running"));
        goto cleanup;
    }

    if (!(disk = virDomainDiskByName(vm->def, path, false))) {
        virReportError(VIR_ERR_INVALID_ARG,
                       _("invalid path: %s"), path);
        goto cleanup;
    }

    ret = vzctlDomainDiskResize(vm, disk->src->path, size);

 cleanup:
    vzctDomainObjEndJob(vm, &job);
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainGetJobInfoImpl(virDomainObjPtr vm, virDomainJobInfoPtr info)
{
    vzctDomainObjPrivatePtr priv = vm->privateData;
    unsigned long long now;

    memset(info, 0, sizeof(*info));

    if (!priv->job || !priv->migration)
        return 0;

    if (virTimeMillisNow(&now) < 0)
        return -1;

    info->type = VIR_DOMAIN_JOB_UNBOUNDED;
    info->dataTotal = 100;
    info->dataProcessed = 0;
    info->dataRemaining = 100;
    info->timeElapsed = now - priv->jobStarted;

    return 0;
}

static int
vzctDomainGetJobInfo(virDomainPtr dom, virDomainJobInfoPtr info)
{
    virDomainObjPtr vm;
    int ret = -1;

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainGetJobInfoEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    ret = vzctDomainGetJobInfoImpl(vm, info);

 cleanup:
    virDomainObjEndAPI(&vm);
    return ret;
}

static int
vzctDomainJobInfoToParams(virDomainJobInfoPtr info,
                          int *type,
                          virTypedParameterPtr *params,
                          int *nparams)
{
    virTypedParameterPtr par = NULL;
    int maxpar = 0;
    int npar = 0;

    if (virTypedParamsAddULLong(&par, &npar, &maxpar,
                                VIR_DOMAIN_JOB_TIME_ELAPSED,
                                info->timeElapsed) < 0 ||
        virTypedParamsAddULLong(&par, &npar, &maxpar,
                                VIR_DOMAIN_JOB_DATA_TOTAL,
                                info->dataTotal) < 0 ||
        virTypedParamsAddULLong(&par, &npar, &maxpar,
                                VIR_DOMAIN_JOB_DATA_PROCESSED,
                                info->dataProcessed) < 0 ||
        virTypedParamsAddULLong(&par, &npar, &maxpar,
                                VIR_DOMAIN_JOB_DATA_REMAINING,
                                info->dataRemaining) < 0)
        goto error;


    *type = info->type;
    *params = par;
    *nparams = npar;
    return 0;

 error:
    virTypedParamsFree(par, npar);
    return -1;
}

static int
vzctDomainGetJobStats(virDomainPtr dom,
                      int *type,
                      virTypedParameterPtr *params,
                      int *nparams,
                      unsigned int flags)
{
    virDomainJobInfo info;
    virDomainObjPtr vm;
    int ret = -1;

    virCheckFlags(0, -1);

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainGetJobStatsEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (vzctDomainGetJobInfoImpl(vm, &info) < 0)
        goto cleanup;

    if (info.type == VIR_DOMAIN_JOB_NONE) {
        *type = VIR_DOMAIN_JOB_NONE;
        *params = NULL;
        *nparams = 0;
        ret = 0;
        goto cleanup;
    }

    ret = vzctDomainJobInfoToParams(&info, type, params, nparams);

 cleanup:
    virDomainObjEndAPI(&vm);

    return ret;
}

static int
vzctDomainGetVcpus(virDomainPtr dom,
                   virVcpuInfoPtr info,
                   int maxinfo,
                   unsigned char *cpumaps,
                   int maplen)
{
    virDomainObjPtr vm;
    size_t i;
    int ret = -1;
    int num;

    if (!(vm = vzctDomObjFromDomain(dom)))
        return -1;

    if (virDomainGetVcpusEnsureACL(dom->conn, vm->def) < 0)
        goto cleanup;

    if (!virDomainObjIsActive(vm)) {
        virReportError(VIR_ERR_OPERATION_INVALID, "%s",
                       _("cannot retrieve vcpu information for inactive domain"));
        goto cleanup;
    }

    if (maxinfo == 0) {
        ret = 0;
        goto cleanup;
    }

    num = virDomainDefGetVcpus(vm->def);
    if (maxinfo < num)
        num = maxinfo;

    if (info) {
        unsigned long long ns;

        memset(info, 0, sizeof(*info) * maxinfo);

        if (vzctlDomainVCPUStats(vm, &ns) < 0)
            goto cleanup;
        ns /= virDomainDefGetVcpus(vm->def);

        for (i = 0; i < num; i++) {
            info[i].number = i;
            info[i].state = VIR_VCPU_RUNNING;
            info[i].cpuTime = ns;
        }
    }

    if (cpumaps) {
        virBitmapPtr bitmap = NULL;
        virBitmapPtr cpumask;
        int hostcpus;

        memset(cpumaps, 0, sizeof(*cpumaps) * maxinfo);

        if (vm->def->cpumask) {
            cpumask = vm->def->cpumask;
        } else {
            if ((hostcpus = virHostCPUGetCount()) < 0)
                goto cleanup;

            if (!(bitmap = virBitmapNew(hostcpus)))
                goto cleanup;

            virBitmapSetAll(bitmap);
            cpumask = bitmap;
        }

        for (i = 0; i < num; i++)
            virBitmapToDataBuf(cpumask, VIR_GET_CPUMAP(cpumaps, maplen, i), maplen);

        virBitmapFree(bitmap);
    }

    ret = num;

 cleanup:
    virDomainObjEndAPI(&vm);

    return ret;
}

static virHypervisorDriver vzctHypervisorDriver = {
    .name = "vzct",
    .connectOpen = vzctConnectOpen, /* 3.2.1 */
    .connectClose = vzctConnectClose, /* 3.2.1 */
    .connectGetVersion = vzctConnectGetVersion, /* 3.2.1 */
    .connectGetHostname = vzctConnectGetHostname, /* 3.2.1 */
    .connectGetSysinfo = vzctConnectGetSysinfo, /* 3.2.1 */
    .connectGetMaxVcpus = vzctConnectGetMaxVcpus, /* 3.2.1 */
    .connectGetCapabilities = vzctConnectGetCapabilities, /* 3.2.1 */
    .connectIsEncrypted = vzctConnectIsEncrypted, /* 3.2.1 */
    .connectIsSecure = vzctConnectIsSecure, /* 3.2.1 */
    .connectIsAlive = vzctConnectIsAlive, /* 3.2.1 */
    .connectListDomains = vzctConnectListDomains, /* 3.2.1 */
    .connectListDefinedDomains = vzctConnectListDefinedDomains, /* 3.2.1 */
    .connectListAllDomains = vzctConnectListAllDomains, /* 3.2.1 */
    .connectNumOfDomains = vzctConnectNumOfDomains, /* 3.2.1 */
    .connectNumOfDefinedDomains = vzctConnectNumOfDefinedDomains, /* 3.2.1 */
    .connectDomainEventRegisterAny = vzctConnectDomainEventRegisterAny, /* 3.2.1 */
    .connectDomainEventDeregisterAny = vzctConnectDomainEventDeregisterAny, /* 3.2.1 */
    .connectGetCPUModelNames = vzctConnectGetCPUModelNames, /* 3.2.1 */
    .connectGetType = vzctConnectGetType, /* 3.2.1 */
    .connectSupportsFeature = vzctConnectSupportsFeature, /* 3.2.1 */

    .nodeGetInfo = vzctNodeGetInfo, /* 3.2.1 */
    .nodeGetCPUStats = vzctNodeGetCPUStats, /* 3.2.1 */
    .nodeGetMemoryStats = vzctNodeGetMemoryStats, /* 3.2.1 */
    .nodeGetCellsFreeMemory = vzctNodeGetCellsFreeMemory, /* 3.2.1 */
    .nodeGetFreeMemory = vzctNodeGetFreeMemory, /* 3.2.1 */
    .nodeGetCPUMap = vzctNodeGetCPUMap, /* 3.2.1 */

    .domainDefineXML = vzctDomainDefineXML, /* 3.2.1 */
    .domainDefineXMLFlags = vzctDomainDefineXMLFlags, /* 3.2.1 */
    .domainGetXMLDesc = vzctDomainGetXMLDesc, /* 3.2.1 */
    .domainUndefine = vzctDomainUndefine, /* 3.2.1 */
    .domainUndefineFlags = vzctDomainUndefineFlags, /* 3.2.1 */
    .domainLookupByID = vzctDomainLookupByID, /* 3.2.1 */
    .domainLookupByUUID = vzctDomainLookupByUUID, /* 3.2.1 */
    .domainLookupByName = vzctDomainLookupByName, /* 3.2.1 */
    .domainIsActive = vzctDomainIsActive, /* 3.2.1 */
    .domainIsPersistent = vzctDomainIsPersistent, /* 3.2.1 */
    .domainIsUpdated = vzctDomainIsUpdated, /* 3.2.1 */
    .domainGetInfo = vzctDomainGetInfo, /* 3.2.1 */
    .domainGetState = vzctDomainGetState, /* 3.2.1 */
    .domainCreate = vzctDomainCreate, /* 3.2.1 */
    .domainCreateWithFlags = vzctDomainCreateWithFlags, /* 3.2.1 */
    .domainDestroy = vzctDomainDestroy, /* 3.2.1 */
    .domainDestroyFlags = vzctDomainDestroyFlags, /* 3.2.1 */
    .domainSuspend = vzctDomainSuspend, /* 3.2.1 */
    .domainResume = vzctDomainResume, /* 3.2.1 */
    .domainBlockStats = vzctDomainBlockStats, /* 3.2.1 */
    .domainBlockStatsFlags = vzctDomainBlockStatsFlags, /* 3.2.1 */
    .domainMemoryStats = vzctDomainMemoryStats, /* 3.2.1 */
    .domainInterfaceStats = vzctDomainInterfaceStats, /* 3.2.1 */
    .domainAttachDevice = vzctDomainAttachDevice, /* 3.2.1 */
    .domainAttachDeviceFlags = vzctDomainAttachDeviceFlags, /* 3.2.1 */
    .domainUpdateDeviceFlags = vzctDomainUpdateDeviceFlags, /* 3.2.1 */
    .domainDetachDevice = vzctDomainDetachDevice, /* 3.2.1 */
    .domainDetachDeviceFlags = vzctDomainDetachDeviceFlags, /* 3.2.1 */
    .domainShutdown = vzctDomainShutdown, /* 3.2.1 */
    .domainShutdownFlags = vzctDomainShutdownFlags, /* 3.2.1 */
    .domainReboot = vzctDomainReboot, /* 3.2.1 */
    .domainReset = vzctDomainReset, /* 3.2.1 */
    .domainGetAutostart = vzctDomainGetAutostart, /* 3.2.1 */
    .domainSetAutostart = vzctDomainSetAutostart, /* 3.2.1 */
    .domainGetOSType = vzctDomainGetOSType, /* 3.2.1 */
    .domainGetVcpusFlags = vzctDomainGetVcpusFlags, /* 3.2.1 */
    .domainGetMaxVcpus = vzctDomainGetMaxVcpus, /* 3.2.1 */
    .domainGetMaxMemory = vzctDomainGetMaxMemory, /* 3.2.1 */
    .domainSetMaxMemory = vzctDomainSetMaxMemory, /* 3.2.1 */
    .domainSetVcpus = vzctDomainSetVcpus, /* 3.2.1 */
    .domainSetVcpusFlags = vzctDomainSetVcpusFlags, /* 3.2.1 */
    .domainSetMemory = vzctDomainSetMemory, /* 3.2.1 */
    .domainSetMemoryFlags = vzctDomainSetMemoryFlags, /* 3.2.1 */
    .domainHasManagedSaveImage = vzctDomainHasManagedSaveImage, /* 3.2.1 */
    .domainManagedSaveRemove = vzctDomainManagedSaveRemove, /* 3.2.1 */
    .domainManagedSave = vzctDomainManagedSave, /* 3.2.1 */
    .domainSnapshotCreateXML = vzctDomainSnapshotCreateXML, /* 3.2.1 */
    .domainSnapshotGetXMLDesc = vzctDomainSnapshotGetXMLDesc, /* 3.2.1 */
    .domainSnapshotNum = vzctDomainSnapshotNum, /* 3.2.1 */
    .domainSnapshotListNames = vzctDomainSnapshotListNames, /* 3.2.1 */
    .domainListAllSnapshots = vzctDomainListAllSnapshots, /* 3.2.1 */
    .domainSnapshotNumChildren = vzctDomainSnapshotNumChildren, /* 3.2.1 */
    .domainSnapshotListChildrenNames = vzctDomainSnapshotListChildrenNames, /* 3.2.1 */
    .domainSnapshotListAllChildren = vzctDomainSnapshotListAllChildren, /* 3.2.1 */
    .domainSnapshotLookupByName = vzctDomainSnapshotLookupByName, /* 3.2.1 */
    .domainHasCurrentSnapshot = vzctDomainHasCurrentSnapshot, /* 3.2.1 */
    .domainSnapshotGetParent = vzctDomainSnapshotGetParent, /* 3.2.1 */
    .domainSnapshotCurrent = vzctDomainSnapshotCurrent, /* 3.2.1 */
    .domainSnapshotIsCurrent = vzctDomainSnapshotIsCurrent, /* 3.2.1 */
    .domainSnapshotHasMetadata = vzctDomainSnapshotHasMetadata, /* 3.2.1 */
    .domainRevertToSnapshot = vzctDomainRevertToSnapshot, /* 3.2.1 */
    .domainSnapshotDelete = vzctDomainSnapshotDelete, /* 3.2.1 */
    .domainMigrateBegin3Params = vzctDomainMigrateBegin3Params, /* 3.2.1 */
    .domainMigratePrepare3Params = vzctDomainMigratePrepare3Params, /* 3.2.1 */
    .domainMigratePerform3Params = vzctDomainMigratePerform3Params, /* 3.2.1 */
    .domainMigrateFinish3Params = vzctDomainMigrateFinish3Params, /* 3.2.1 */
    .domainMigrateConfirm3Params = vzctDomainMigrateConfirm3Params, /* 3.2.1 */
    .domainAbortJob = vzctDomainAbortJob, /* 3.2.1 */
    .domainSetUserPassword = vzctDomainSetUserPassword, /* 3.2.1 */
    .domainBlockResize = vzctDomainBlockResize, /* 3.2.1 */
    .domainGetJobInfo = vzctDomainGetJobInfo, /* 3.2.1 */
    .domainGetJobStats = vzctDomainGetJobStats, /* 3.2.1 */
    .domainGetVcpus = vzctDomainGetVcpus, /* 3.2.1 */
};

static virConnectDriver vzctConnectDriver = {
    .hypervisorDriver = &vzctHypervisorDriver,
};

static virStateDriver vzctStateDriver = {
    .name = "vzct",
    .stateInitialize = vzctStateInitialize,
    .stateCleanup = vzctStateCleanup,
    .stateAutoStart = vzctStateAutoStart,
};

int
vzctRegister(void)
{
    if (virRegisterConnectDriver(&vzctConnectDriver, true) < 0)
        return -1;

    if (virRegisterStateDriver(&vzctStateDriver) < 0)
        return -1;

    return 0;
}
