/*
 * vzct_vzctl.c: core driver functions for managing
 * Virtuozzo Containers
 *
 * Copyright (C) 2017 Virtuozzo, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>
#include <signal.h>
#include <fcntl.h>

#include <vzctl/libvzctl.h>
#include <ploop/libploop.h>

#include "virlog.h"
#include "virfile.h"
#include "viralloc.h"
#include "virstring.h"
#include "virpidfile.h"
#include "virprocess.h"
#include "vircommand.h"

#include "vzct_utils.h"
#include "vzct_vzctl.h"

#define VIR_FROM_THIS VIR_FROM_VZCT

VIR_LOG_INIT("vzct.vzct_vzctl");

#define CHECK(label, func, ...) do { \
    int __ret__; \
    if ((__ret__ = func(__VA_ARGS__))) { \
        virReportError(VIR_ERR_INTERNAL_ERROR, _("%s returned %d"), #func,  __ret__); \
        goto label; \
    } \
} while (0)

#define IGNORE(func, ...) do { \
    int __ret__; \
    if ((__ret__ = func(__VA_ARGS__))) \
        VIR_WARN("%s returned %d", #func,  __ret__); \
} while (0)

static vzctl_env_handle_ptr
vzctlEnvOpen(virDomainObjPtr dom,
             const char *filename, const char *funcname, size_t linenum)
{
    int err;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    vzctl_env_handle_ptr env;

    virUUIDFormat(dom->def->uuid, uuidstr);
    if (!(env = vzctl2_env_open(uuidstr, 0, &err))) {
        virReportErrorHelper(VIR_FROM_THIS, VIR_ERR_INTERNAL_ERROR,
                             filename, funcname, linenum,
                             _("vzctl2_env_open error: %d"), err);
        return NULL;
    }

    return env;
}

#define ENV_OPEN_RET(ret) \
    do { \
        if (!(env = vzctlEnvOpen(dom, __FILE__, __FUNCTION__, __LINE__))) \
            return ret; \
    } while (0)

#define ENV_OPEN \
    ENV_OPEN_RET(-1)

#define ENV_CLOSE  vzctl2_env_close(env)

static int vzctlSetParams(virDomainObjPtr dom, vzctl_env_handle_ptr env);
static void vzctlDetachDisks(vzctl_env_handle_ptr env);
static char* vzctlPloopGetTopName(const char *path);

int
vzctlDomainCreate(virDomainObjPtr dom, bool paused, char *dumppath)
{
    int ret = -1;
    vzctl_env_handle_ptr env = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    vzctDomainObjPrivatePtr priv = dom->privateData;
    struct vzctl_env_create_param create_param = {
        .ostmpl = NULL,
        /* FIXME */
        .config = (char*) "vswap.1024MB",
        .ve_private = NULL,
        .ve_root = NULL,
        .layout = VZCTL_LAYOUT_5,
        .no_root_disk = 1,
        .enc_keyid = NULL,
    };
    vzctl_env_param_ptr env_param;

    if ((env_param = vzctl2_alloc_env_param()) == NULL) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Can't allocate vzctl handle."));
        return -1;
    }

    virUUIDFormat(dom->def->uuid, uuidstr);
    create_param.uuid = uuidstr;
    if (!virStrcpy(create_param.ctid, uuidstr, sizeof(create_param.ctid))) {
        virReportError(VIR_ERR_INTERNAL_ERROR,
                       _("Can't copy uuid, dest size if %lu, src is %lu"),
                       sizeof(create_param.ctid), strlen(uuidstr) + 1);
        return -1;
    }
    create_param.name = dom->def->name;

    CHECK(cleanup, vzctl2_env_create, env_param, &create_param, 0);
    priv->created = true;
    ENV_OPEN;

    if (vzctlSetParams(dom, env) < 0)
        goto cleanup;

    ENV_CLOSE;
    ENV_OPEN;

    if (!paused) {
        if (dumppath) {
            struct vzctl_cpt_param cpt = {
                .cmd = VZCTL_CMD_RESTORE,
                .dumpfile = (char*)dumppath /* FIXME */
            };
            CHECK(cleanup, vzctl2_env_restore, env, &cpt, 0);
        } else {
            CHECK(cleanup, vzctl2_env_start, env, 0);
        }
    }

    ret = 0;

 cleanup:
    /* leave container partially created for a moment */
    ENV_CLOSE;
    return ret;
}

static int
vzctlAddDisk(vzctl_env_handle_ptr handle,
             virDomainDiskDefPtr disk,
             bool root)
{
    struct vzctl_disk_param params = { .enabled = 0 };

    params.enabled = 1;
    params.size = 0;
    params.path = disk->src->path;
    if (root)
        params.mnt = (char*) "/"; /* FIXME */

    if (disk->src->type == VIR_STORAGE_TYPE_BLOCK)
        params.use_device = 1;

    CHECK(error, vzctl2_env_attach_disk, handle, &params);

    return 0;
 error:
    return -1;
}

static int
vzctlAddFs(vzctl_env_handle_ptr handle, virDomainFSDefPtr fs)
{
    struct vzctl_disk_param params = { .enabled = 0 };

    params.enabled = 1;
    params.size = 0;
    params.path = fs->src->path;
    params.mnt = fs->dst;

    if (fs->type == VIR_DOMAIN_FS_TYPE_BLOCK)
        params.use_device = 1;

    CHECK(error, vzctl2_env_attach_disk, handle, &params);

    return 0;
 error:
    return -1;
}

static int
vzctlDeleteDisk(vzctl_env_handle_ptr handle, virDomainDiskDefPtr disk)
{
    vzctl_env_param_ptr env_param = vzctl2_get_env_param(handle);
    struct vzctl_disk *it = NULL;
    struct vzctl_disk_param disk_param;

    while ((it = vzctl2_env_get_disk(env_param, it))) {
        CHECK(error, vzctl2_env_get_disk_param, it, &disk_param, sizeof(disk_param));
        if (STREQ_NULLABLE(disk_param.path, disk->src->path)) {
            CHECK(error, vzctl2_env_detach_disk, handle, disk_param.uuid);
            return 0;
        }
    }

    virReportError(VIR_ERR_INTERNAL_ERROR,
                   _("Can't find disk with source '%s'."), disk->src->path);
 error:
    return -1;
}

static int
vzctlDeleteFs(vzctl_env_handle_ptr handle, virDomainFSDefPtr fs)
{
    vzctl_env_param_ptr env_param = vzctl2_get_env_param(handle);
    struct vzctl_disk *it = NULL;
    struct vzctl_disk_param disk;

    while ((it = vzctl2_env_get_disk(env_param, it))) {
        CHECK(error, vzctl2_env_get_disk_param, it, &disk, sizeof(disk));
        if (STREQ_NULLABLE(disk.mnt, fs->dst)) {
            CHECK(error, vzctl2_env_detach_disk, handle, disk.uuid);
            return 0;
        }
    }

    virReportError(VIR_ERR_INTERNAL_ERROR,
                   _("Can't find disk with mnt '%s'."), fs->dst);
 error:
    return -1;
}

static int
vzctBootDiskIndex(virDomainDefPtr def, int *ret)
{
    size_t i;
    int bootDisk = -1;
    bool hasRootFs = false;

    for (i = 0; i < def->ndisks; i++) {
        int idx = def->disks[i]->info.bootIndex;

        if (idx && (bootDisk == -1 || idx < def->disks[bootDisk]->info.bootIndex))
            bootDisk = i;
    }

    for (i = 0; i < def->nfss; i++)
        if (STREQ_NULLABLE(def->fss[i]->dst, "/"))
            hasRootFs = true;

    if (hasRootFs && bootDisk != -1) {
        virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                       _("Root disk should be specified either by "
                         "root filesystem or by disk boot order not both."));
        return -1;
    }

    if (!hasRootFs && bootDisk == -1) {
        if (def->ndisks == 0) {
            virReportError(VIR_ERR_CONFIG_UNSUPPORTED, "%s",
                           _("root disk is not specified"));
            return -1;
        }
        bootDisk = 0;
    }

    *ret = bootDisk;
    return 0;
}

static int
vzctlConfigureDisks(vzctl_env_handle_ptr handle, virDomainDefPtr def)
{
    size_t i;
    int ret = -1;
    int bootDisk;

    if (vzctBootDiskIndex(def, &bootDisk) < 0)
        return -1;

    for (i = 0; i < def->ndisks; i++)
        if (vzctlAddDisk(handle, def->disks[i], i == bootDisk) < 0)
            goto cleanup;

    for (i = 0; i < def->nfss; i++) {
        if (vzctlAddFs(handle, def->fss[i]) < 0)
            goto cleanup;
    }

    ret = 0;
 cleanup:

    return ret;
}

static void
vzctlDetachDisks(vzctl_env_handle_ptr env)
{
    vzctl_env_param_ptr env_param = vzctl2_get_env_param(env);
    struct vzctl_disk *it = NULL;
    struct vzctl_disk_param disk;

    while ((it = vzctl2_env_get_disk(env_param, NULL))) {
        IGNORE(vzctl2_env_get_disk_param, it, &disk, sizeof(disk));
        IGNORE(vzctl2_env_detach_disk, env, disk.uuid);
    }
}

static int
vzctlAddNet(vzctl_env_param_ptr env_param, virDomainNetDefPtr net)
{
    vzctl_veth_dev_iterator veth_it = NULL;
    struct vzctl_veth_dev_param vethParam;
    char macString[VIR_MAC_STRING_BUFLEN];

    memset(&vethParam, 0, sizeof(vethParam));

    virMacAddrFormat(&net->mac, macString);
    vethParam.mac_ve = macString;

    vethParam.dev_name = net->ifname;
    vethParam.dev_name_ve = net->ifname_guest;
    vethParam.dhcp = 2;

    if (!(veth_it = vzctl2_create_veth_dev(&vethParam, sizeof(vethParam)))) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Can't create veth device"));
        goto error;
    }

    CHECK(error, vzctl2_env_add_veth, env_param, veth_it);
    return 0;

 error:
    if (veth_it)
        vzctl2_free_veth_dev(veth_it);
    return -1;
}

static int
vzctlDeleteNet(vzctl_env_param_ptr env_param, virDomainNetDefPtr net)
{

    CHECK(error, vzctl2_env_del_veth, env_param, net->ifname_guest);
    return 0;

 error:
    return -1;
}

static int
vzctlConfigureNets(vzctl_env_param_ptr env_param, virDomainDefPtr def)
{
    size_t i;

    IGNORE(vzctl2_env_del_veth, env_param, "all");

    for (i = 0; i < def->nnets; i++) {
        if (vzctlAddNet(env_param, def->nets[i]) < 0)
            goto error;
    }

    return 0;

 error:
    return -1;
}

static int
vzctlSetParams(virDomainObjPtr dom, vzctl_env_handle_ptr env)
{
    vzctl_env_param_ptr env_param;
    int ret = -1;
    char *cpumask = NULL;

    if ((env_param = vzctl2_alloc_env_param()) == NULL) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Can't allocate vzctl handle."));
        return -1;
    }

    CHECK(cleanup, vzctl2_env_set_ramsize,
          env_param, dom->def->mem.cur_balloon >> 10);

    CHECK(cleanup, vzctl2_env_set_cpu_count,
          env_param, virDomainDefGetVcpus(dom->def));

    if (dom->def->cpumask) {
        cpumask = virBitmapFormat(dom->def->cpumask);
        CHECK(cleanup, vzctl2_env_set_cpumask, env_param, cpumask);
    }

    vzctlDetachDisks(env);

    if (vzctlConfigureDisks(env, dom->def) < 0)
        goto cleanup;

    if (vzctlConfigureNets(env_param, dom->def) < 0)
        goto cleanup;

    CHECK(cleanup, vzctl2_apply_param, env,
          env_param, VZCTL_SKIP_SETUP|VZCTL_SAVE);

    ret = 0;

 cleanup:
    VIR_FREE(cpumask);
    vzctl2_free_env_param(env_param);
    return ret;
}

bool
vzctlDomainActive(virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    vzctl_env_handle_ptr env = NULL;
    vzctl_env_status_t status;
    bool ret = false;

    ENV_OPEN_RET(false);
    CHECK(cleanup, vzctl2_get_env_status_info, env, &status, ENV_STATUS_ALL);
    priv->created = true;
    ret = !!(status.mask & ENV_STATUS_RUNNING);

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainDestroy(virDomainObjPtr dom, stop_mode_e stop_mode)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    if (!priv->created)
        return 0;

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_stop, env, stop_mode, 0);
    priv->created = false;
    vzctlDetachDisks(env);
    CHECK(cleanup, vzctl2_env_destroy, env, 0);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainReset(virDomainObjPtr dom)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_stop, env, M_KILL_FORCE, 0);
    CHECK(cleanup, vzctl2_env_start, env, 0);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainPause(virDomainObjPtr dom)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_pause, env, 0);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainUnpause(virDomainObjPtr dom)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_start, env, 0);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainRestart(virDomainObjPtr dom)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_restart, env, 0);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainSuspend(virDomainObjPtr dom,
                   const char *dumppath)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;
    struct vzctl_cpt_param cpt = {
        .cmd = VZCTL_CMD_CHKPNT,
        .dumpfile = (char*)dumppath /* FIXME */
    };

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_chkpnt, env, VZCTL_CMD_CHKPNT, &cpt, 0);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

static int
vzctlDomainDiskGetUUID(vzctl_env_handle_ptr env,
                       const char *path,
                       char *uuid)
{
    struct vzctl_disk *it = NULL;
    struct vzctl_disk_param disk;
    vzctl_env_param_ptr env_param = vzctl2_get_env_param(env);

    while ((it = vzctl2_env_get_disk(env_param, it))) {
        CHECK(error, vzctl2_env_get_disk_param, it, &disk, sizeof(disk));
        if (STREQ(disk.path, path)) {
            memcpy(uuid, disk.uuid, VIR_UUID_STRING_BUFLEN + 2);
            return 0;
        }
    }

    virReportError(VIR_ERR_INTERNAL_ERROR,
                   _("cannot find disk uuid for path: %s"), path);

 error:
    return -1;
}

int
vzctlDomainVCPUStats(virDomainObjPtr dom, unsigned long long *ns)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;
    struct vzctl_cpustat s;

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_cpustat, env, &s, sizeof(s));
    /* convert seconds to nanoseconds by the way */
    *ns = (s.user + s.system + s.nice) * (1000 * 1000 * 1000);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainBlockStats(virDomainObjPtr dom,
                      const char *path,
                      virDomainBlockStatsPtr stats)
{
    char uuid[VIR_UUID_STRING_BUFLEN + 2];
    struct vzctl_disk_stats vzstats;
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    if (vzctlDomainDiskGetUUID(env, path, uuid) < 0)
        goto cleanup;

    CHECK(cleanup, vzctl2_env_get_disk_stats, env, uuid, &vzstats,
                                            sizeof(vzstats));

    stats->rd_bytes = vzstats.io.read;
    stats->wr_bytes = vzstats.io.write;
    stats->rd_req = -1;
    stats->wr_req = -1;
    stats->errs = -1;

    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainBlockStatsAll(virDomainObjPtr dom,
                         virDomainBlockStatsPtr stats)
{
    vzctl_env_param_ptr env_param;
    struct vzctl_disk *it = NULL;
    struct vzctl_disk_param disk;
    struct vzctl_disk_stats vzstats;
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    env_param = vzctl2_get_env_param(env);
    stats->rd_bytes = 0;
    stats->wr_bytes = 0;

    while ((it = vzctl2_env_get_disk(env_param, it))) {
        CHECK(cleanup, vzctl2_env_get_disk_param, it, &disk, sizeof(disk));
        CHECK(cleanup, vzctl2_env_get_disk_stats, env, disk.uuid,
                                                &vzstats, sizeof(vzstats));

        stats->rd_bytes += vzstats.io.read;
        stats->wr_bytes += vzstats.io.write;
    }

    stats->rd_req = -1;
    stats->wr_req = -1;
    stats->errs = -1;

    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDomainMemoryStats(virDomainObjPtr dom,
                       virDomainMemoryStatPtr stats,
                       unsigned int nr_stats)
{
    char uuid[VIR_UUID_STRING_BUFLEN];
    struct vzctl_meminfo mem;
    size_t i = 0;

    virUUIDFormat(dom->def->uuid, uuid);
    CHECK(error, vzctl2_get_env_meminfo, uuid, &mem, sizeof(mem));

#define SET_STAT(TAG, VALUE)            \
    if (i < nr_stats) {                 \
        stats[i].tag = (TAG);           \
        stats[i].val = (VALUE);         \
        i++;                            \
    }

    /* bytes to kB */
    SET_STAT(VIR_DOMAIN_MEMORY_STAT_SWAP_IN, mem.swap_in)
    SET_STAT(VIR_DOMAIN_MEMORY_STAT_SWAP_OUT, mem.swap_out)
    SET_STAT(VIR_DOMAIN_MEMORY_STAT_AVAILABLE, mem.total >> 10)
    SET_STAT(VIR_DOMAIN_MEMORY_STAT_ACTUAL_BALLOON, mem.total >> 10)
    SET_STAT(VIR_DOMAIN_MEMORY_STAT_UNUSED, mem.free >> 10)

#undef SET_STAT

    return i;

 error:
    return -1;
}

int vzctlAttachDisk(virDomainObjPtr dom, virDomainDeviceDefPtr dev)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    if (dev->type == VIR_DOMAIN_DEVICE_DISK) {
        ret = vzctlAddDisk(env, dev->data.disk, false);
    } else if (dev->type == VIR_DOMAIN_DEVICE_FS) {
        ret = vzctlAddFs(env, dev->data.fs);
    } else {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("unsupported device type"));
        goto cleanup;
    }

 cleanup:
    ENV_CLOSE;
    return ret;
}

int vzctlDetachDisk(virDomainObjPtr dom, virDomainDeviceDefPtr dev)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;

    ENV_OPEN;
    if (dev->type == VIR_DOMAIN_DEVICE_DISK) {
        ret = vzctlDeleteDisk(env, dev->data.disk);
    } else if (dev->type == VIR_DOMAIN_DEVICE_FS) {
        ret = vzctlDeleteFs(env, dev->data.fs);
    } else {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("unsupported device type"));
        goto cleanup;
    }

    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int vzctlAttachNet(virDomainObjPtr dom, virDomainNetDefPtr net)
{
    vzctl_env_handle_ptr env = NULL;
    vzctl_env_param_ptr env_param = NULL;
    int ret = -1;

    ENV_OPEN;
    if ((env_param = vzctl2_alloc_env_param()) == NULL) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Can't allocate vzctl handle."));
        goto cleanup;
    }

    if (vzctlAddNet(env_param, net) < 0)
        goto cleanup;

    CHECK(cleanup, vzctl2_apply_param, env, env_param, VZCTL_SAVE);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    vzctl2_free_env_param(env_param);
    return ret;
}

int vzctlDetachNet(virDomainObjPtr dom, virDomainNetDefPtr net)
{
    vzctl_env_handle_ptr env = NULL;
    vzctl_env_param_ptr env_param = NULL;
    int ret = -1;

    ENV_OPEN;
    if ((env_param = vzctl2_alloc_env_param()) == NULL) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Can't allocate vzctl handle."));
        goto cleanup;
    }

    if (vzctlDeleteNet(env_param, net) < 0)
        goto cleanup;

    CHECK(cleanup, vzctl2_apply_param, env, env_param, VZCTL_SAVE);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    vzctl2_free_env_param(env_param);
    return ret;
}

int
vzctlSetVcpus(virDomainObjPtr dom,
              unsigned int nvcpus)
{
    vzctl_env_handle_ptr env = NULL;
    vzctl_env_param_ptr env_param;
    int ret = -1;

    ENV_OPEN;
    if ((env_param = vzctl2_alloc_env_param()) == NULL) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Can't allocate vzctl handle."));
        goto cleanup;
    }

    CHECK(cleanup, vzctl2_env_set_cpu_count, env_param, nvcpus);
    CHECK(cleanup, vzctl2_apply_param, env, env_param, VZCTL_SAVE);

    ret = 0;

 cleanup:
    vzctl2_free_env_param(env_param);
    ENV_CLOSE;
    return ret;
}

int
vzctlSetMemory(virDomainObjPtr dom,
               unsigned long newmem)
{
    vzctl_env_handle_ptr env = NULL;
    vzctl_env_param_ptr env_param;
    int ret = -1;

    ENV_OPEN;
    if ((env_param = vzctl2_alloc_env_param()) == NULL) {
        virReportError(VIR_ERR_INTERNAL_ERROR, "%s",
                       _("Can't allocate vzctl handle."));
        goto cleanup;
    }

    CHECK(cleanup, vzctl2_env_set_ramsize, env_param, newmem >> 10);
    CHECK(cleanup, vzctl2_apply_param, env, env_param, VZCTL_SAVE);

    ret = 0;

 cleanup:
    vzctl2_free_env_param(env_param);
    ENV_CLOSE;
    return ret;
}

virDomainSnapshotObjListPtr
vzctlLoadSnapshots(virDomainObjPtr dom)
{
    vzctl_env_handle_ptr env = NULL;
    vzctl_env_param_ptr env_param;
    virDomainSnapshotObjListPtr snapshots = NULL;
    virDomainSnapshotDefPtr def = NULL;
    virDomainSnapshotObjPtr snapshot;
    struct vzctl_snapshot_tree *tree = NULL;
    const char *private;
    char *snapshotsXmlPath;
    int err;
    size_t i;

    if (!(snapshots = virDomainSnapshotObjListNew()))
        return NULL;

    ENV_OPEN_RET(NULL);
    env_param = vzctl2_get_env_param(env);

    CHECK(cleanup, vzctl2_env_get_ve_private_path, env_param, &private);

    if (!(snapshotsXmlPath = virFileBuildPath(private, "Snapshots.xml", NULL)))
        goto cleanup;

    tree = vzctl2_open_snapshot_tree(snapshotsXmlPath, &err);
    if (err) {
        virReportError(VIR_ERR_INTERNAL_ERROR, _("Can't open snapshots: %d"), err);
        goto cleanup;
    }

    VIR_INFO("Number of snapshots is %d", tree->nsnapshots);
    for (i = 0; i < tree->nsnapshots; i++) {
        struct vzctl_snapshot_data *s = tree->snapshots[i];

        if (VIR_ALLOC(def) < 0)
            goto cleanup;

        if (VIR_STRDUP(def->name,  s->guid) < 0 ||
            VIR_STRDUP(def->parent, s->parent_guid) < 0 ||
            VIR_STRDUP(def->description, s->name) < 0)
            goto cleanup;
        def->memory = VIR_DOMAIN_SNAPSHOT_LOCATION_NONE;
        def->state = VIR_DOMAIN_SHUTOFF;
        def->current = s->current;

        if (!(snapshot = virDomainSnapshotAssignDef(snapshots, def)))
            goto cleanup;

        def = NULL;

        VIR_INFO("%d, %s, %s, %s", s->current, s->guid, s->parent_guid, s->name);
    }


 cleanup:
    if (tree)
        vzctl2_close_snapshot_tree(tree);
    ENV_CLOSE;
    return snapshots;
}

static char *
vzctlGetSnapGuid(vzctl_env_handle_ptr env, const char *name)
{
    struct vzctl_snapshot_tree *tree = NULL;
    vzctl_env_param_ptr env_param;
    const char *private;
    char *snapshotsXmlPath;
    int err;
    size_t i;
    char *ret = NULL;

    env_param = vzctl2_get_env_param(env);
    CHECK(cleanup, vzctl2_env_get_ve_private_path, env_param, &private);

    if (!(snapshotsXmlPath = virFileBuildPath(private, "Snapshots.xml", NULL)))
        goto cleanup;

    tree = vzctl2_open_snapshot_tree(snapshotsXmlPath, &err);
    if (err) {
        virReportError(VIR_ERR_INTERNAL_ERROR, _("Can't open snapshots: %d"), err);
        goto cleanup;
    }

    VIR_INFO("Number of snapshots is %d", tree->nsnapshots);
    for (i = 0; i < tree->nsnapshots; i++) {
        struct vzctl_snapshot_data *s = tree->snapshots[i];

        if STREQ_NULLABLE(name, s->name) {
            ignore_value(VIR_STRDUP(ret, s->guid));
            goto cleanup;
        }
    }

 cleanup:
    if (tree)
        vzctl2_close_snapshot_tree(tree);
    return ret;
}

int
vzctlCreateSnapshot(virDomainObjPtr dom, virDomainSnapshotObjPtr snap)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;
    struct vzctl_snapshot_param param = {
        .name = snap->def->name,
        .desc = snap->def->description,
    };

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_create_snapshot, env, &param);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlRevertSnapshot(virDomainObjPtr dom,
                    virDomainSnapshotObjPtr snap ATTRIBUTE_UNUSED)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;
    char *guid = NULL;
    struct vzctl_switch_snapshot_param param = { .flags = 0 };

    ENV_OPEN;
    if (!(guid = vzctlGetSnapGuid(env, snap->def->name)))
        goto cleanup;

    param.guid = guid;

    CHECK(cleanup, vzctl2_env_switch_snapshot, env, &param);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

int
vzctlDeleteSnapshot(virDomainObjPtr dom,
                    virDomainSnapshotObjPtr snap ATTRIBUTE_UNUSED)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;
    char *guid = NULL;

    ENV_OPEN;
    if (!(guid = vzctlGetSnapGuid(env, snap->def->name)))
        goto cleanup;

    CHECK(cleanup, vzctl2_env_delete_snapshot, env, guid);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

static bool
vzctMigrationDiskCopy(const char *disk,
                      size_t nmigrate_disks,
                      const char **migrate_disks)
{
    size_t i;

    if (!nmigrate_disks)
        return true;

    for (i = 0; i < nmigrate_disks; i++) {
        if (STREQ(disk, migrate_disks[i]))
            return true;
    }

    return false;
}

static int
vzctlMigrationAddServerDisk(virCommandPtr cmd,
                            const char *disk,
                            int nmigrate_disks,
                            const char **migrate_disks)
{
    char *image;

    if (!vzctMigrationDiskCopy(disk, nmigrate_disks, migrate_disks))
        return 0;

    if (!(image = vzctlPloopGetTopName(disk)))
        return -1;

    virCommandAddArgList(cmd, "--image", image, NULL);

    VIR_FREE(image);
    return 0;
}

int
vzctlMigrationStartServer(vzctDriverPtr driver,
                          virDomainObjPtr dom,
                          int nmigrate_disks,
                          const char **migrate_disks,
                          bool copy)
{
    int ret = -1;
    virCommandPtr cmd = NULL;
    vzctDomainObjPrivatePtr priv = dom->privateData;
    unsigned short port = 0;
    size_t i;
    int logfd = -1;

    if (virPortAllocatorAcquire(driver->migrationPorts, &port) < 0)
        goto cleanup;
    priv->migrationPort = port;

    if ((logfd = open("/var/log/phaul-service.log",
                      O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR)) < 0) {
        virReportSystemError(errno, "%s", _("cannot open /var/log/phaul-service.log"));
        goto cleanup;
    }

    cmd = virCommandNew("/usr/libexec/vzct-migration-dst.py");

    virCommandSetOutputFD(cmd, &logfd);
    virCommandSetErrorFD(cmd, &logfd);
    virCommandAddArgList(cmd, "0.0.0.0", NULL);
    virCommandAddArgFormat(cmd, "%d", priv->migrationPort);

    if (copy) {
        for (i = 0; i < dom->def->ndisks; ++i)
            if (vzctlMigrationAddServerDisk(cmd, dom->def->disks[i]->src->path,
                                            nmigrate_disks, migrate_disks) < 0)
                goto cleanup;

        for (i = 0; i < dom->def->nfss; ++i)
            if (vzctlMigrationAddServerDisk(cmd, dom->def->fss[i]->src->path,
                                            nmigrate_disks, migrate_disks) < 0)
                goto cleanup;
    }

    if (virCommandRunAsync(cmd, &priv->migrationPid))
        goto cleanup;

    ret = 0;

 cleanup:
    VIR_FORCE_CLOSE(logfd);
    virCommandFree(cmd);

    return ret;
}

void
vzctlMigrationStopServer(vzctDriverPtr driver ATTRIBUTE_UNUSED, virDomainObjPtr dom)
{
    vzctDomainObjPrivatePtr priv = dom->privateData;

    virProcessKill(priv->migrationPid, SIGTERM);
    ignore_value(virProcessWait(priv->migrationPid, NULL, false));
    priv->migrationPid = -1;
    virPortAllocatorRelease(priv->migrationPort);
    priv->migrationPort = 0;
}

static char*
vzctlPloopGetTopName(const char *path)
{
    struct ploop_disk_images_data *dd = NULL;
    char *ddpath;
    char *image = NULL;
    char *buf = NULL;

    if (virAsprintf(&ddpath, "%s/DiskDescriptor.xml", path) < 0)
        return NULL;

    if (VIR_ALLOC_N(buf, 256) < 0)
        goto cleanup;

    CHECK(cleanup, ploop_open_dd, &dd, ddpath);

    CHECK(cleanup, ploop_get_top_delta_fname, dd, buf, 256);

    if (VIR_STRDUP(image, buf) < 0)
        goto cleanup;

 cleanup:
    VIR_FREE(ddpath);
    VIR_FREE(buf);
    ploop_close_dd(dd);

    return image;
}

static int
vzctlMigrationAddClientDisk(virCommandPtr cmd,
                            vzctl_env_handle_ptr env,
                            const char *disk,
                            const char *mount,
                            int nmigrate_disks,
                            const char **migrate_disks,
                            bool copy)
{
    char *image;
    char uuid[VIR_UUID_STRING_BUFLEN + 2];
    const char *disktype = "shared";

    if (!(image = vzctlPloopGetTopName(disk)))
        return -1;

    if (vzctlDomainDiskGetUUID(env, disk, uuid) < 0)
        goto cleanup;

    if (copy && vzctMigrationDiskCopy(disk, nmigrate_disks, migrate_disks))
        disktype = "local";

    virCommandAddArgList(cmd, "--disk", image, uuid, mount, disktype, NULL);

 cleanup:
    VIR_FREE(image);
    return 0;
}

int
vzctlMigrationStartClient(virDomainObjPtr dom, const char *uristr, int port,
                          int nmigrate_disks,
                          const char **migrate_disks,
                          bool copy,
                          bool live)
{
    int ret = -1;
    virCommandPtr cmd = NULL;
    vzctl_env_handle_ptr env = NULL;
    char uuidstr[VIR_UUID_STRING_BUFLEN];
    char *shared = NULL;
    virURIPtr uri = NULL;
    size_t i;
    vzctDomainObjPrivatePtr priv = dom->privateData;
    int bootDisk;
    int logfd = -1;

    ENV_OPEN;
    if (!(uri = virURIParse(uristr)))
        goto cleanup;

    if ((logfd = open("/var/log/phaul.log",
                      O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR)) < 0) {
        virReportSystemError(errno, "%s", _("cannot open /var/log/phaul.log"));
        goto cleanup;
    }

    cmd = virCommandNew("/usr/libexec/vzct-migration-src.py");

    virCommandSetOutputFD(cmd, &logfd);
    virCommandSetErrorFD(cmd, &logfd);
    virUUIDFormat(dom->def->uuid, uuidstr);
    virCommandAddArgList(cmd, uri->server, NULL);
    virCommandAddArgFormat(cmd, "%d", port);
    virCommandAddArgList(cmd, uuidstr, NULL);

    if (live)
        virCommandAddArgList(cmd, "--live", NULL);

    vzctBootDiskIndex(dom->def, &bootDisk);

    for (i = 0; i < dom->def->ndisks; ++i)
        if (vzctlMigrationAddClientDisk(cmd, env,
                                        dom->def->disks[i]->src->path,
                                        i == bootDisk ? "/" : "",
                                        nmigrate_disks, migrate_disks, copy) < 0)
            goto cleanup;

    for (i = 0; i < dom->def->nfss; ++i)
        if (vzctlMigrationAddClientDisk(cmd, env,
                                        dom->def->fss[i]->src->path,
                                        dom->def->fss[i]->dst,
                                        nmigrate_disks, migrate_disks, copy) < 0)
            goto cleanup;

    if (virCommandRunAsync(cmd, &priv->migrationPid) < 0)
        goto cleanup;

    ret = 0;

 cleanup:
    VIR_FORCE_CLOSE(logfd);
    ENV_CLOSE;
    virURIFree(uri);
    virCommandFree(cmd);
    VIR_FREE(shared);

    return ret;
}

int vzctlDomainSetUserPassword(virDomainObjPtr dom,
                               const char *user,
                               const char *password,
                               unsigned int flags)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;
    int f = 0;

    if (flags & VIR_DOMAIN_PASSWORD_ENCRYPTED)
        f |= VZCTL_SET_USERPASSWD_CRYPTED;

    ENV_OPEN;
    CHECK(cleanup, vzctl2_env_set_userpasswd, env, user, password, f);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}

/* size in KiB */
int vzctlDomainDiskResize(virDomainObjPtr dom,
                          const char *path,
                          unsigned long long size)
{
    vzctl_env_handle_ptr env = NULL;
    int ret = -1;
    char uuid[VIR_UUID_STRING_BUFLEN + 2];

    ENV_OPEN;
    if (vzctlDomainDiskGetUUID(env, path, uuid) < 0)
        goto cleanup;

    CHECK(cleanup, vzctl2_env_resize_disk, env, uuid, size, false);
    ret = 0;

 cleanup:
    ENV_CLOSE;
    return ret;
}
